#! /usr/bin/env python
# Copyright (C) 2009 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Main module responsible for backing up files"""

# standard library imports
import gettext
import datetime as dt
import os.path
import time
import threading
from sqlite3 import OperationalError
# >= python 2.6 Queue.PriorityQueue exists. We added for
# future backwards compatibility right now
try:
    from Queue import Queue, PriorityQueue
    # The Result Queue: for each assigned task, what happened with it. This
    # Queue has no priority.
    ReQueue = Queue.Queue()
    class PriorityQueue(Queue.PriorityQueue):

        def _put(self, item, heappush=heapq.heappush):
            heappush(self.queue, item)
            ReQueue.put(item[1])

except ImportError:
    import heapq
    import Queue

    # The Result Queue: for each assigned task, what happened with it. This
    # Queue has no priority.
    ReQueue = Queue.Queue()
    class PriorityQueue(Queue.Queue):
        '''Variant of Queue that retrieves open entries in priority order
        (lowest first).

        Entries are typically tuples of the form:  (priority number, data)
        This class can be found at: Python-2.6a3/Lib/Queue.py
        '''

        maxsize = 0

        def _init(self, maxsize):
            self.queue = []

        def _qsize(self, len=len):
            return len(self.queue)

        def _put(self, item, heappush=heapq.heappush):
            heappush(self.queue, item)
            ReQueue.put(item[1])

        def _get(self, heappop=heapq.heappop):
            return heappop(self.queue)

# related 3rd party imports
from bzrlib import transport

# local application/ library specific imports.
from myrrule import *
import createdb as cdb
import task as t
import watcher as w
from logger import *
# TODO: we should use a lazy importer here in order to prevent importing all
# future backups methods
from bum_rdiff import backup

# connects with the database
sqlcon = cdb.SQLConnection(timeout_=20)
_tb = t.TemporalBasis()
# establish translation context
gettext.bindtextdomain('Ghosts', 'locale')
gettext.textdomain('Ghosts')
_ = gettext.gettext

#__all__ = ['TaskScheduler']

# the TaskScheduler Queue: the priority is assigned by the date of the next
# backup
TSQueue = PriorityQueue()
# the FileUpdater Queue: the priority is assigned by
# TemporalBasis.teb_priority
FUQueue = PriorityQueue()
# The Signaling Queue: a simple queue to pass an alert message to all
# threads (e.g. to stop activities)
SigQueue = PriorityQueue()

def _task_useful_data(task):
    """From a given task instance returns only the attributes we need to
    pass onto a job instance"""
    priority = sqlcon.find(t.TemporalBasis, ("teb_priority",),
                          dict(teb_id=task.tas_teb_id))[0]['teb_priority']
    return dict(job_tas_id=task.tas_id, job_from=task.tas_from,
                job_including=task.tas_including,
                job_excluding=task.tas_excluding,
                job_sto_id=task.tas_sto_id, job_to=task.tas_to,
                job_priority=priority, job_timestamp=int(time.time()),
                job_scheduled_date=dt.datetime.now())

def _seconds_from_epoch(cmp_val):
    """Returns the difference between cmp_val and now (in seconds from
    epoch)"""
    return int(cmp_val-time.time())

def _fromT2dt(time_):
    """Converts a time.localtime() tuple into datetime.datetime (only applies
    year, month, day, hour, minute and second)"""
    t = time.localtime(time_)
    return dt.datetime(t[0], t[1], t[2], t[3], t[4], t[5])

def _fromdt2T(dt):
    """Converts a datetime.datetime into seconds since epoch (only applies
    year, month, day, hour, minute and seconds)"""
    return time.mktime((dt.year, dt.month, dt.day, dt.hour, dt.minute,
            dt.second, 0, 0, 0))

class JobError(Exception):
    """Handler for all errors related to Job"""
    pass

class TaskScheduler(object):
    """Manages the creation of new schedules for an existing task (only when
    Temporal_Basis.teb_validate is True)
    """

    tasks = {} # the live instances of every task
    dates = {} # for every task a list of scheduled dates.
    blacklisted = [] # tasks that have been cancelled after the daemon started
                     # working.

    def __init__(self):
        """Will retrieve last init timestamp from coordinator_special_event or
        insert it in case it does not exist"""
        sqlcon.cursor.execute("""select cse_timestamp from
        coordinator_special_event where cse_action = 'init'""")
        try:
            self.init = sqlcon.cursor.fetchone()[0]
        except TypeError:
            self.init = dt.datetime.now()
            sqlcon.cursor.execute("""insert into coordinator_special_event
            (cse_action, cse_timestamp) values ('init', ?)""", (self.init,))
            sqlcon.commit()
        self.epoch = _fromdt2T(self.init)

    def __update_db(self):
        """Updates the init value for coordinator_special_event in order to
        persist the last calculated date"""
        self.init = dt.datetime.now()
        sqlcon.cursor.execute("""update coordinator_special_event set
        cse_timestamp=? where cse_action='init'""", (self.init,))
        sqlcon.commit()
        self.epoch = _fromdt2T(self.init)

    def __iter_dates(self):
        """Return a tuple/dict of (task, next_scheduled_date)"""
        for every_task in self.dates.iteritems():
            try:
                for every_date in every_task[1]:
                    yield (every_task[0], every_date)
            except TypeError, errmsg:
                if not every_task[1]:
                    continue
                else:
                    log.error(errmsg)

    def __dates_values(self):
        """Return an iterator on every date if date is smaller than now or -if
        it's bigger- the difference is less than 5 minutes"""
        for fst in self.dates.values():
            if fst is not None:
                for snd in fst:
                    # TODO: should find a smart way to establish a tolerable
                    # elapsed time: probably should keep a moving average
                    # window of last jobs
                    if snd<dt.datetime.now() or \
                    (_fromdt2T(snd)-time.time())<=300:
                        yield snd

    def __task_lookup(self, dates):
        """Given scheduled dates, it will look up in every member of the tasks
        dict and yield all tasks where the dates exist in a tuple of (task
        instance, date)"""
        for fst in self.dates.iteritems():
            for date in dates:
                try:
                    if date in (fst[1]):
                        self.dates[fst[0]].remove(date)
                        yield (self.tasks[fst[0]], date)
                except TypeError, errmsg:
                    if not fst[1]:
                        continue
                    else:
                        log.error(errmsg)

    def __eliminate_done(self):
        """Given a list of scheduled dates, erases the ones that has been
        already accomplished"""
        sqlcon.cursor.execute("""create temporary table if not exists
        temp_to_erase (tas_id int, scheduled_date timestamp)""")
        sqlcon.cursor.execute("""create unique index if not exists temp_idx_job
        on temp_to_erase (tas_id, scheduled_date)""")
        sqlcon.cursor.executemany("""insert or replace into temp_to_erase
        (tas_id, scheduled_date) values (?, ?)""", self.__iter_dates())
        sqlcon.cursor.execute("""select distinct job_tas_id, job_scheduled_date
        from job inner join temp_to_erase on (job_tas_id=tas_id and
        job_scheduled_date=scheduled_date) inner join job_detail on
        (job_id=jod_job_id) where jod_log = 'File successfully backuped'""")
        rc = sqlcon.cursor.fetchall()
        for row in rc:
            self.dates[row[0]].remove(row[1])

    def next_scheduled_task(self):
        """Given a populated instance, it will search for the next task(s)
        -a list of tasks in case several of them share the same date-"""
        next_task = list()
        try:
            min_date = list(self.__dates_values())
            min_date.sort()
        except ValueError:
            return None
        next_task += [val for val in self.__task_lookup(min_date)]
        return next_task

    def populate(self, how_many=None):
        """Will query the database and bring all active tasks
        If parameter how_many contains an integer it will calculate the next
        "how_many" scheduled dates for every task"""
        valid_tebs = sqlcon.find(_tb, ("teb_id", ), dict(teb_validate=True))
        # these are brand new tasks (not because the tasks have just been
        # created but because this might be the first incarnation of the class)
        # Here we  want all active tasks whose temporal_basis.teb_validate is
        # true (typically this puts out 'in the moment' and 'on demand')
        rc = sqlcon.find(t.Task, None, dict(tas_cancel_date=None,
        tas_teb_id=map(lambda s: s["teb_id"], valid_tebs)), True)
        self.tasks = dict([(r.tas_id, r) for r in rc])
        if how_many:
            for cur in self.tasks.itervalues():
                try:
                    self.dates[cur.tas_id] = \
                            self.next_scheduled_date(
                               cur, self.init, how_many)
                except TypeError, errmsg:
                    if errmsg.message == "iteration over non-sequence":
                        pass
                    else:
                        log.error(errmsg)
        self.__eliminate_done()
        self.__update_db()

    def calculate(self, how_many=1):
        """Given a list of dicts populated with task information, will return
        for each of them the next "how_many" scheduled dates (seconds since
        epoch)"""
        for cur in self.tasks.itervalues():
            if self.dates.has_key(cur.tas_id):
                del self.dates[cur.tas_id][:]
            self.dates[cur.tas_id] = self.next_scheduled_date(cur, self.init,
                                                              how_many)

    @staticmethod
    def next_scheduled_date(cls, last_date=None, how_many=1):
        """Given a date (datetime.now() in case None is defined)
        return the next scheduled date(s) for a given task. If the
        task's TemporalBasis does not need an scheduled data (e.g. On
        Demand) it will return None.

        Parameters:
            cls: a task.Task instance
            last_date: a datetime instance, it will be use to calculate the
            next date after it
            how_many: how many days after the compared date this method must
            retrieve.
        """

        if cls.tas_time_coord is None:
            return None

        # aux. var. we will use to compare.
        if last_date is None:
        # if it's not defined, we get now()
            cmp_dt = dt.datetime.now()
        else:
        # else, we copy
            cmp_dt = last_date

        # hacking HD to prevent KeyErrors on several scenarios (this should
        # be cover in task.validate() but in case it's not executed, this
        # give a little additional check)
        if not cls.tas_time_coord or not cls.tas_time_coord.has_key('HD'):
            cls.tas_time_coord['HD'] = [dt.time(0,0,0)]

        # special case: selected monthly backup but no day and no hour:
        # just last_day shortcut. If that's the case, we deal with it
        # right now and break from this method without further work.
        if cls.tas_teb_id == _tb.MONTHLY and \
           cls.tas_time_coord.has_key('LD') and \
           cls.tas_time_coord['LD'] and \
           (not cls.tas_time_coord.has_key('DM') or \
           not cls.tas_time_coord['DM']):
            return list(rrule(MONTHLY, dtstart=cmp_dt, count=how_many,
                              bymonthday=1, byhour=0, byminute=0))

        # we use the appropiate func according to temporal_basis
        if cls.tas_teb_id == _tb.MONTHLY:
            if not cls.tas_time_coord.has_key('DM'): return None
            # did the user selected a "last_day" backup?
            if cls.tas_time_coord.has_key('LD') and \
               cls.tas_time_coord['LD']:
                rrule_list = rruleset()
                # standard rule for common days
                rrule_lis1 = rrule(MONTHLY, cache=True,
                                   dtstart=cmp_dt,
                                   count=how_many,
                                   bymonthday=cls.tas_time_coord['DM'],
                                   bytime=cls.tas_time_coord['HD'])
                # rule for last_day shortcut
                rrule_lis2 = rrule(MONTHLY, cache=True,
                                   dtstart=cmp_dt.replace(hour=0,
                                   minute=0), count=how_many,
                                   bymonthday=1)
                rrule_list.rrule(rrule_lis1)
                rrule_list.rrule(rrule_lis2)

                return [val for (idx,val) in enumerate(rrule_list) \
                                                   if idx < how_many]

            else:
                return list(rrule(MONTHLY, cache=True,
                            dtstart=cmp_dt,
                            count=how_many,
                            bymonthday=cls.tas_time_coord['DM'],
                            bytime=cls.tas_time_coord['HD']))

        elif cls.tas_teb_id == _tb.WEEKLY:
            if not cls.tas_time_coord.has_key('DW'): return None
            return list(rrule(WEEKLY, cache=True,
                        dtstart=cmp_dt,
                        count=how_many,
                        byweekday=cls.tas_time_coord['DW'],
                        bytime=cls.tas_time_coord['HD']))

        elif cls.tas_teb_id == _tb.DAILY:
            if not cls.tas_time_coord.has_key('HD'): return None
            return list(rrule(DAILY, cache=True,
                        dtstart=cmp_dt,
                        count=how_many,
                        bytime=cls.tas_time_coord['HD']))

        elif cls.tas_teb_id == _tb.HOURLY:
            if not cls.tas_time_coord.has_key('HD'): return None
            return list(rrule(HOURLY, cache=True,
                        dtstart=cmp_dt, count=how_many,
                        byminute=[val.minute for val in \
                                  cls.tas_time_coord['HD']]))

class Job(object):
    """
    The object used to communicate files to update between queues.
    This objects holds all the data we need to know to perform the task:
    job_id: the job identification number
    job_tas_id: the task identification number
    job_from: the path to the file
    jod_file: the name of the file
    job_sto_id: the transport instance needed to transfer the backup'ed file
    job_to: the directory where to put the file
    job_priority: the value needed in the FUQueue Priority Queue to priorize or
                  not the job.
    job_timestamp: the timestamp when this task was scheduled
    jod_timestamp: the timestamp when a specific file from the task was scheduled
    jod_waf_id: the watched file identification number
    jod_backup_date: the date when the backup was performed.
    jod_backup_name: the name assigned to the file in the backup repository
    jod_log: a message with the final result (success, failure, etc)
    """

    __table__ = "job"
    __pk__ = "job_id"
    __references__ = "job_detail"
    __fk__ = "jod_job_id"
    job_id = None
    TS_attrs = set(('job_tas_id', 'job_from', 'job_including', 'job_excluding',
                    'job_sto_id', 'job_to', 'job_priority', 'job_scheduled_date',
                    'job_timestamp'))
    FU_attrs = set(('jod_path', 'jod_file', 'jod_timestamp', 'job_tas_id',
                    'job_timestamp', 'job_sto_id', 'job_to', 'jod_waf_id',
                    'job_priority'))
    BU_attrs = set(('jod_path', 'jod_file', 'jod_timestamp', 'jod_backup_date',
                    'jod_backup_name', 'jod_log'))

    class dummy(object):
        __table__ = "job_detail"

    def __init__(self, id=None, data=None):
        """Data must always be a list of dicts with the attribute:value as
        explained in the class doc string. Only job_id will be ignored since it
        will be assigned by the db. In case id is assigned, it will be used to
        update the job info in the db (e.g. after a backup job was performed)
        """
        if id:
            if type(id) == list:
                self.jobs = []
                for every in id:
                    self.jobs.append(sqlcon.get(self, every,
                        **dict(job_id=job_id, job_tas_id=job_tas_id,
                            job_from=job_from, jod_file=jod_file,
                            jod_path=jod_path, job_including=job_including,
                            job_excluding=job_excluding,
                            job_sto_id=job_sto_id, job_to=job_to,
                            job_priority=job_priority,
                            job_scheduled_date=job_scheduled_date,
                            job_timestamp=job_timestamp,
                            jod_backup_date=job_backup_date,
                            jod_backup_name=job_backup_name, jod_log=job_log,
                            jod_timestamp=jod_timestamp
                            )))
            elif type(id) == int:
                rc = sqlcon.get(self, id,
                        **dict(job_id=job_id, job_tas_id=job_tas_id,
                            job_from=job_from, jod_file=jod_file,
                            jod_path=jod_path, job_including=job_including,
                            job_excluding=job_excluding,
                            job_sto_id=job_sto_id, job_to=job_to,
                            job_priority=job_priority,
                            job_scheduled_date=job_scheduled_date,
                            job_timestamp=job_timestamp,
                            jod_backup_date=job_backup_date,
                            jod_backup_name=job_backup_name, jod_log=job_log,
                            jod_timestamp=jod_timestamp
                            ))
                for x in rc.items():
                    setattr(self, x[0], x[1])
        elif data:
            if type(data) == dict:
                for x in data.items():
                    setattr(self, x[0], x[1])
            elif type(data) == list:
                self.jobs = data

    def job_distribution(self):
        """This method reads the event, finds out the job fp and derivates the
        job to the given activity"""
        TSData = []
        FUData = []
        BUData = []
        if hasattr(self, 'jobs'):
            for every in self.jobs:
                if self.TS_attrs.issubset(every.keys()):
                    TSData.append(every)
                elif self.FU_attrs.issubset(every.keys()):
                    FUData.append(every)
                elif self.BU_attrs.issubset(every.keys()):
                    BUData.append(every)
        else:
            if self.TS_attrs.issubset(self.__dict__.keys()):
                TSData.append(self.__dict__)
            elif self.FU_attrs.issubset(self.__dict__.keys()):
                FUData.append(self.__dict__)
            elif self.BU_attrs.issubset(self.__dict__.keys()):
                BUData.append(self.__dict__)

        if TSData:
            self.TS_Data(TSData)
        if FUData:
            self.FU_Data(FUData)
        if BUData:
            self.BU_Data(BUData)

    def TS_Data(self, values):
        """Use this method to insert fresh TaskScheduler info (see
        self.TS_attrs for the attributes required)"""
        sqlcon.insert(self, values, 'abort', True)

    def FU_Data(self, values):
        """Use this method to insert fresh FileUpdater info (see
        self.FU_attrs for the attributes required)"""
        pass_list = []
        for counter, every in enumerate(values):
            sqlcon.cursor.execute("""select job_id from job where job_tas_id=?
            and job_timestamp=?""", (every['job_tas_id'],
            every['job_timestamp']))
            values[counter]['jod_job_id'] = sqlcon.cursor.fetchone()[0]
            pass_list.append(dict([cur for cur in every.items() \
                    if cur[0].startswith("jod_")]))
        sqlcon.insert(self.dummy, pass_list, 'abort', True)

    def BU_Data(self, values):
        """Use this method to insert fresh BackUp info (see self.BU_attrs for
        the attributes required)"""
        sqlcon.update(self.dummy, dict([cur for cur in values[0].items() if \
                                        cur[0] in ('jod_backup_date',
                                        'jod_backup_name', 'jod_log')]),
                                  dict([cur for cur in values[0].items() if \
                                       cur[0] in ('jod_path', 'jod_file',
                                       'jod_timestamp')]), True)

class BUThread(threading.Thread):
    """Thread(s) responsible for performing the backup. As many as threads are
    defined will be triggered by Coordinator: they will read from FUQueue,
    backup the file to the desired Transport and update the Job database table
    with the result of the operation"""

    def __init__(self, Tname, sleep_interval):
        self.sleep_int = sleep_interval
        threading.Thread.__init__(self, name=Tname)

    def run(self):
        """Read BUThread doc string"""
        while 1:
            if self.__stop():
                break
            # get a new file to backup
            while not FUQueue.empty():
                pr, jdb = FUQueue.get_nowait()
                bckp_date = dt.datetime.now()
                backup(os.path.join(jdb.jod_path,jdb.jod_file),
                       "%d_%d_%d.diff" % (jdb.jod_tas_id, jdb.jod_waf_id,
                       _fromdt2T(bckp_date)), jdb.jod_to, 'zip')
                # Additionally, update Job info
                ReQueue.put(Job(None, dict(jod_path=jdb.jod_path,
                    jod_file=jdb.jod_file, jod_timestamp=jdb.jod_timestamp,
                    jod_backup_date=bckp_date, jod_backup_name="%d_%d_%d"
                    % (jdb.job_tas_id, jdb.jod_waf_id, _fromdt2T(bckp_date)),
                    jod_log='File successfully backuped')))
                # check again if we have to stop
                if self.__stop():
                    break
            time.sleep(self.sleep_int)

    def __stop(self, return_type=False):
        try:
            # first see if there's a signal to stop
            pr, val = SigQueue.get_nowait()
        except Queue.Empty:
            if return_type:
                return False, None
            else:
                return False
        if val.has_key('abort') or val.has_key('clean'):
            SigQueue.put_nowait((-1, val))
            if return_type:
                if val.has_key('abort'):
                    return True, 'abort'
                elif val.has_key('clean'):
                    return True, 'clean'
                else:
                    return True, None
            else:
                return True
        else:
            if return_type:
                return False, None
            else:
                return False

class FUThread(threading.Thread):
    """Thread responsible for receiving the tasks scheduled and performing the
    track files update (meaning updating to see if new files has been added in
    the tracked paths and returning a list of all files that has changed since
    last backup"""

    def __init__(self, Tname, sleep_interval):
        self.sleep_int = sleep_interval
        threading.Thread.__init__(self, name=Tname)

    def run(self):
        """Loop ad infinitum unless it gets a stop signal. Periodically check
        TSQueue and find out if a new task has been queued. If so, update its
        tracked files info to discover new files and changed files. Put them
        on FUQueue and sleep before starting all over again"""
        FUsqlcon = cdb.SQLConnection(20)
        while 1:
            while not TSQueue.empty():
                pr, jdb = TSQueue.get_nowait()
                if self.__stop():
                    FUsqlcon.close()
                    break
                wf = t.WatchFile(jdb.job_tas_id, FUsqlcon)
                # first we need to update the tree to see if new files were
                # added
                wf.track_file(jdb.job_from, jdb.job_including,
                        jdb.job_excluding, True, jdb.job_tas_id, False)
                FUsqlcon.commit()
                if self.__stop():
                    FUsqlcon.close()
                    break
                # find out which has changed
                # TODO: make self.__stop() signal a break instruction from
                # within the function: callback?
                changed_files = wf.get_changed_files(self.__stop())
                FUsqlcon.commit()
                # Additionally, update Job info
                for every in changed_files:
                    FU_dict = dict(jod_path=every['waf_path'],
                            jod_file=every['waf_name'],
                            jod_waf_id=every['waf_id'],
                            job_priority=jdb.job_priority,
                            job_sto_id=jdb.job_sto_id,
                            job_to=jdb.job_to,
                            jod_timestamp=int(time.time()),
                            job_tas_id=jdb.job_tas_id,
                            job_timestamp=jdb.job_timestamp)
                    qjob = Job(None, FU_dict)
                    FUQueue.put((qjob.job_priority, qjob))
                # check again if we have to stop
                if self.__stop():
                    FUsqlcon.close()
                    break
            time.sleep(self.sleep_int)

    def __stop(self):
        # first see if there's a signal to stop
        try:
            pr, val = SigQueue.get_nowait()
        except Queue.Empty:
            return False
        if val.has_key('abort') or val.has_key('clean'):
            SigQueue.put_nowait((-1, val))
            return True
        else:
            return False

class Coordinator(object):
    """
    Responsible of the main control of the backup process, this class will
    loop receiving in each iteration a possible new file to update and sending
    it to the backups threads. Additionally it coordinates other threads such
    as the TaskScheduler and the FileUpdater.
    """

    def __init__(self):
        gp = t.GlobalParameter()
        self.max_threads = gp.get_global_maximum_threads()[0]
        self.sle_interval = gp.get_global_sleep_interval()[0]
        self.tem_storage = gp.get_global_temporary_storage()[0]
        self.Threads = []
        sqlcon.cursor.execute("""delete from coordinator_special_event""")
        sqlcon.commit()
        self.init_time = time.time()

    def loop(self, callbacks=[]):
        """
        The main function of this class: will iterate for ever coordinating
        the work of all Threads. Accept a list of callbacks functions
        (executing them in FIFO style)
        """
        while 1:
            if self._stop():
                # perform some final cleanup if we were asked to stop.
                break
            # if defined, execute the callbacks functions
            if callbacks:
                for cur in callbacks:
                    cur(self)
            if not self.Threads:
                # we create the FileUpdater thread: in charge of calculating
                # the new existing files and passing them to the BackUp
                # Thread.
                new_thread = FUThread("FileUpdaterThread", self.sle_interval)
                new_thread.start()
                self.Threads.append(new_thread)
                # we create the BackUp Thread(s) in charge of the proper work
                # of backup'ing the files. We know how many threads we need to
                # generate using the
                # task.GlobalParameter.get_global_maximum_threads method
                for cur in range(1, self.max_threads+1):
                    new_thread = BUThread("BackupThread%s" % (cur),
                            self.sle_interval)
                    new_thread.start()
                    self.Threads.append(new_thread)
                # here we get all inotify events to backup, passing the, directly
                # to BUThread.
                NEv = w.NEvent(None)
                # finally in the startup we instance TaskScheduler which is
                # responsible for calculating the next tasks to be backup'ed
                TS = TaskScheduler()
            # we should get inotify events a.s.a.p. and pass them to the
            # FUQueue, we cannot wait for FUThread to be processed.
            # Note to myself: it will be faster to place this action directly
            # in the Backup thread, however it needs to communicate with the
            # database and sqlite is not thread safe. That's why it will stay
            # here
            rc = sqlcon.find(NEv, {"min(eve_seq)":"min_eve_seq",
                                   "max(eve_seq)":"max_eve_seq"}, None, False)
            if rc[0]['min_eve_seq'] and rc[0]['max_eve_seq']:
                min_max_eve = (rc[0]['min_eve_seq'], rc[0]['max_eve_seq'])
                nevents = NEv.pop(True, *min_max_eve)
                self.process_events(nevents)
                del nevents
            # here we calculate the next scheduled dates to perform backups
            # and load the information in the FileUpdater
            try:
                TS.populate(10)
                next_tasks = TS.next_scheduled_task()
            except OperationalError, errmsg:
                if errmsg == "database is locked":
                    time.sleep(self.sle_interval)
            for every in next_tasks:
                try:
                    ta = _task_useful_data(every[0])
                except OperationalError, errmsg:
                    if errmsg == "database is locked":
                        time.sleep(self.sle_interval)
                qjob = Job(None, ta)
                TSQueue.put((qjob.job_priority, qjob))
            # we get the status of every job and we save them in the database
            # right now
            while not ReQueue.empty():
                jdb = ReQueue.get_nowait()
                try:
                    jdb.job_distribution()
                except OperationalError, errmsg:
                    if errmsg == "database is locked":
                        time.sleep(self.sle_interval)
            # let's take a little nap if defined: good night!
            time.sleep(self.sle_interval)

    def stop(self, stop_type='clean', sqlcon_=None):
        """Ask the coordinator to stop every thread, all jobs and exit the
        loop.
        Parameters:
                stop_type: accept two values: clean (all current jobs will
                finish before stopping) and abort (jobs will be stop abruptly
                in whatever state they are before stopping). This method won't
                stop from logging the result of all already finished jobs
        """
        if sqlcon_:
            mysqlcon = sqlcon_
        else:
            mysqlcon = sqlcon
        if stop_type in ('clean', 'abort'):
            mysqlcon.cursor.execute("""insert into coordinator_special_event
            (cse_action, cse_timestamp) values (?, ?)""", (stop_type,
            dt.datetime.now()))
            mysqlcon.commit()

    def _stop(self):
        """Perform a cleanup of pending jobs"""
        sqlcon.cursor.execute("""select cse_action from
        coordinator_special_event where cse_action in ('clean', 'abort') and
        cse_timestamp>=?""", (_fromT2dt(self.init_time),))
        try:
            row = sqlcon.cursor.fetchone()[0]
        except TypeError:
            return False
        SigQueue.put_nowait((-1, dict(row=True)))
        # TODO: write a cleanup method to see if the threads are still running
        # TODO: should add join?
        if row == 'clean':
            pass
        elif row == 'abort':
            pass
        # log out all pending job information
        while not ReQueue.empty():
            jdb = ReQueue.get_nowait()
            jdb.job_distribution()
        sqlcon.cursor.execute("""delete from coordinator_special_event""")
        sqlcon.commit()
        return True

    def process_events(self, events):
        """
        Process every event received from inotify watcher. This are the steps
        to perform (for every event):
                1- validate the event (althoough watcher should send the right
                events)
                2- instantiate the corresponding Tasks and validate they still
                exist
                3- perform a globster inclusion to verify we need to backup the
                value.
                4- build the Job class and pass it to the FUQueue
        """
        # for every event, we proceed:
        all_list = []
        pass_flag = True
        check_ta_dict = {}
        new_task = False
        for counter, every in enumerate(events):
            pass_flag = True
            new_task = False
            # TODO: validate event
            if not check_ta_dict.has_key(every.eve_tas_id):
                check_ta = t.Task(every.eve_tas_id)
                TS_dict = _task_useful_data(check_ta)
                check_ta_dict[every.eve_tas_id] = TS_dict
                new_task = True
            else:
                TS_dict = check_ta_dict[every.eve_tas_id]
            FU_dict = dict(jod_path=every.eve_path, jod_file=every.eve_name,
                    jod_timestamp=int(time.time()),
                    job_sto_id=TS_dict['job_sto_id'],
                    jod_waf_id=0,
                    job_to=TS_dict['job_to'],
                    job_tas_id=TS_dict['job_tas_id'],
                    job_priority=TS_dict['job_priority'],
                    job_timestamp=TS_dict['job_timestamp'])
            BU_dict = dict(jod_path=every.eve_path, jod_file=every.eve_name,
                    jod_timestamp=FU_dict['jod_timestamp'])
            # validate task:
            # is the task cancelled?
            if check_ta.tas_cancel_date:
                BU_dict['jod_log'] = (_("""Task %d is cancelled since %s""" %
                    (check_ta.tas_id, check_ta.tas_cancel_date)))
                pass_flag = False
            # is the temporal basis still the same?
            if pass_flag and _tb.INTHEMOMENT != check_ta.tas_teb_id:
                BU_dict['jod_log'] = (_("""The selected task is no longer an \
an immediate backup file"""))
                pass_flag = False
            # globster check
            if pass_flag and not check_ta.cluded_file(every.eve_name):
                BU_dict['jod_log'] = (_("""The file name.extension is \
excluded in the task"""))
                pass_flag = False
            # TODO: add other checks?
            if new_task:
                new_event = Job(None, TS_dict)
                ReQueue.put(new_event)
            if pass_flag:
                qjob = Job(None, FU_dict)
                FUQueue.put((qjob.job_priority, qjob))
            else:
                all_list.append(FU_dict)
                all_list.append(BU_dict)
        all_events = Job(None, all_list)
        ReQueue.put(all_events)

if __name__ == '__main__':
    Coord = Coordinator()
    Coord.loop()
