# Copyright (C) 2016 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Rdiff-backup based backup method

This module defines all methods required to successfully backup a file using the
rdiff-backup method.

Expected roadmap:
    * The method works this way:
        * get the signature for a given file that has been modified (of course
          the original unmodified version should be in the target directory)
          using the SigFile method.
        * Apply the modified file signature to the original file to obtain a
          diff version using the DeltaFile method and store the diff file.
        * Whenever we want to get the original file backup, apply the diff to
          modified version: this will return the original version.
        * This method always keep the last version of the file and stores the
          deltas of previous versions so that users can restore the prior
          incarnations of the file whenever they want.
    * To make this method fit our established plan, we need to take in
      consideration the defined transport for the task: if the target directory
      is local, we can do the work without transfering information. In case the
      target is in a remote device or placed in a mounted device we need to
      transfer as little information as posible.
    * A suggested plan could be the following:
        * if local transport:
            * perform the DeltaFile with two file-like objects locally.
            * save the diff file before replacing the original file with the
              modified one: we will have the last version and then all deltas
              that can be use to patch this last version in reverse until user
              gets the version she/he wants.
        * if remote transport:
            * with every last version of a file, we store its signature (of
              course, we're talking about the remote target), compressed.
            * If the file in the local side changed, we connect and request
              signature.
            * Server sends it and we decompressed it.
            * Apply this signature to the modified local version and get a
              Delta File.
            * Compress the Delta, send it to the remote server and discard the
              received signature.
            * Patch the server version with the Delta File (all described until
              now is according to Martin Pool's paper "rdiff 1.0pre DRAFT, 20th
              July 2004") and generate the SigFile and compress it (this
              compressed sigfile will be needed the next time to backup),
              saving a temporal copy of the original version.
            * Apply DeltaFile to the temporal copy of the original file and the
              signature of our new modified version.
            * Get a diff we can now store as an incremental reverse version.
            * Discard the temporal original version.

For more on rdiff-backup please see: http://www.nongnu.org/rdiff-backup/
For more on librsync / rdiff please see: http://librsync.sourcefrog.net/
"""
# standard library imports
import os
import os.path
import shutil
import tempfile

# related 3rd party imports

# local application/ library specific imports.
from rdiff_backup import librsync, Rdiff, rpath, Globals
#from logger import *
from bzrlib import errors as bzr_errors

class BumRdiffError(Exception):
    pass

def __compress(stream, method=None):
    if not method:
        return stream
    else:
        return stream.encode(method)

def __decompress(stream, method=None):
    if not method:
        return stream
    else:
        return stream.decode(method)

def sign(infile, compression_method=None):
    """Given a file-like object will get its signature and return it.
    The object will be closed after calculating the signature
    Parameters:
        infile: the file-like object
        compression_method: the compression algorithm used to return the
        result. It supports all codecs supported by Python. Options are: zip,
        bz2 and None.
    """
    bs = Rdiff.find_blocksize(os.path.getsize(infile.name))
    re = librsync.SigFile(infile, bs)
    while not re.eof:
        yield __compress(re.read(Globals.blocksize), compression_method)

def delta(signature, file_, compression_method=None):
    """Generator to obtain the diff between a file and its counterpart's
    signature. File will be closed when its done
    Parameters:
        file_: the file-like object
        compression_method: the compression algorithm used to return the
        result. It supports all codecs supported by Python. Options are: zip,
        bz2 and None.
    """
    delta = librsync.DeltaFile(signature, file_)
    while not delta.eof:
        yield __compress(delta.read(Globals.blocksize), compression_method)

def patch(basis_file, delta_file):
    """Generator to obtain the results of patching a file"""
    pf = librsync.PatchedFile(basis_file, delta_file)
    while not pf.eof:
        try:
            yield pf.read(Globals.blocksize)
        except librsyncError, errmsg:
            #log.error(errmsg)
            raise BumRdiffError(errmsg)

def backup(file_name, delta_name, transport, compress_method=None):
    """This function will distribute the backup task to a specific method based
    on the transport type: if its a local transport, it will call
    backward_delta and if it's a remote backup it will use forward_delta"""
    if transport.sto_transport == u'file://':
        backward_delta(file_name, delta_name, transport.transport,
                       compress_method)
    elif transport.sto_transport == u'bzr+ssh://':
        forward_delta(file_name, delta_name, transport.transport,
                      compress_method)

def forward_delta(file_name, delta_name, transport, to, compress_method=None):
    """Generate the delta of a file, using the original file's signature and
    the modified version file. This method will:
        * bring the original file signature (osig for short) from the target
          directory (compressed).
        * decompressed it
        * generate the delta (a non-reverse delta) between the osig and the
          modified file (df for short)
        * send the delta to the target directory (td for short) and discard the
          osig
        * in the td make a temp copy of the original file (ofc for short)
        * patch the original file with the df in order to obtain the modified
          version located in the local target
        * discard df
        * now everything will continue as backward_delta method will little
          changes to reflect we're working on a remote server
    """
    fl_name = os.path.basename(file_name)
    # first we bring the file from the remote server:
    try:
        osig = transport.get("%s.sig" % (str(fl_name)))
    except bzr_errors.NoSuchFile:
        # the sig file does not exist in the remote server: either this is a
        # first time backup or the file was erased, we call backward_delta and
        # call it a day
        backward_delta(file_name, delta_name, transport, compress_method)
        return
    # we have the file, let's try to diff
    orig = open(file_name, 'rb')
    deltaf, delnam = tempfile.mkstemp()
    # write the delta
    for every in delta(osig, orig, 'zip'):
        os.write(deltaf, every)
    os.close(deltaf)
    deltaf = open(delnam)
    # take it to its final destination
    try:
        transport.put_file(os.path.basename(delnam), deltaf)
        transport.rename(os.path.basename(delnam), delta_name)
    finally:
        deltaf.close()
        os.remove(delnam)

def backward_delta(file_name, delta_name, transport, compress_method=None):
    """Generate the delta of a file, using the modified file's signature and
    the prior version file. This method will:
        * generate the delta and saving it using delta_name as name of the
          file.
        * replace the original file with the modified version
        * save the compressed signature of the modified file
    Parameters:
        modified: the name of the "from" (a.k.a. to-backup) version of the
        file.
        original: the name of the "to" (a.k.a. to-be-replaced) version of
        the file
        delta_name: the full path to the file we'll use to save the delta we'll
        generate from the last two parameters
        transport: a bzrlib.Transport instance
        compress_method: in case of need user can select a compression method
        to apply to the delta file and to the signature file. The options are
        the ones that Python has supported builtin codecs (see
        http://docs.python.org/lib/standard-encodings.html#standard-encodings)
    """
    fl_name = os.path.basename(file_name)
    no_delta = False
    # first we open the files we will need to read: the two versions of the
    # same file.
    try:
        orig = transport.get(fl_name)
    except (bzr_errors.NoSuchFile, bzr_errors.InvalidURL):
        # the file does not exist: ergo, it's the first time we are going to
        # backup. If that is the case, we just need to create a signature, put
        # it over there and copy the file.
        no_delta = True
    mod = open(file_name, 'rb')
    # let's produce the signature of the newer file since we are going to rdiff
    # it
    sig = ""
    for every in sign(mod, None):
        sig += every
    if not no_delta:
        # now the file that will contain the delta is created
        deltaf, delnam = tempfile.mkstemp()
        # write the delta
        for every in delta(sig, orig, compress_method):
            os.write(deltaf, every)
        os.close(deltaf)
        deltaf = open(delnam)
        # take it to its final destination
        try:
            transport.put_file(os.path.basename(delnam), deltaf)
            transport.rename(os.path.basename(delnam), delta_name)
        finally:
            deltaf.close()
            os.remove(delnam)
    # let's generate a file with the used signature (this has no use at all
    # when the backup server is local but will be useful in case the local
    # server transforms into a remote one)
    sigtmp, signam = tempfile.mkstemp()
    os.write(sigtmp, __compress(sig, compress_method))
    os.close(sigtmp)
    sigtmp = open(signam)
    try:
        transport.put_file(os.path.basename(signam), sigtmp)
        transport.rename(os.path.basename(signam), "%s.sig" % (str(fl_name)))
    finally:
        sigtmp.close()
        os.remove(signam)
    # let's delete the sig, we have no more use for it.
    del sig
    # copy the modified version of the file, replacing the original one.
    mod = open(file_name, 'rb')
    transport.put_file(str(fl_name), mod)
    mod.close()

def remote_backup(modified, original, delta_name, compress_method=None,
        signature=None):
    """FIXME: write doc"""
    # bring the signature:
    or_sig_str = ""
    orig_sig = open(signature, 'rb')
    while True:
        data = orig_sig.read(Globals.blocksize)
        if data == "": break
        or_sig_str += __decompress(data)
    # now, delta the modified file
    deltaf1 = open("%s.diff" % (modified))
    deltaf1.write(delta(sig, orig, compress_method))
    deltaf1.write(or_sig_str)
    deltaf1.close()
    del or_sig_str
    # now use this delta to generate a reverse delta on the other side
    # first make a temporal copy of the file to backup:
    tmp = tempfile.TemporaryFile()
    orig_file = open(original, 'rb')
    shutil.copyfileobj(orig_file, tmp, Globals.blocksize)
    orig_file.close()
    orig_file = open(original, 'wb')
    deltaf1.open("%s.diff" % (modified), 'rb')
    orig_file.write(patch(tmp, __decompress(deltaf1)))
    deltaf1.close()
    os.remove(deltaf1)
    # now delta:
    deltaf2 = open(delta_name, 'wb')
    # WE NEED TO DELTA!
    #deltaf.write(delta(sig, tmp, compress_method))
    #deltaf.close()
    #orig_file.close()
