# -*- coding: utf8 -*-
# Copyright (C) 2008 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""This module includes:

 A set of functions that create the database needed for this project.

  1- All table's names must be compound by singular words.
     Right: scheduled_task
     Wrong: scheduled_tasks
  2- All tables' names lowercase, underscore separating words.
  3- All columns from a table must follow these guidelines:
     i.   xxx_column_name where xxx stands for an identification for the
          table: if the table name has only one word, the first three
          letters. if the table name has two words, the first two from the
          first word and the the first one from the second word. If a table
          has three words in its name, the first letter from each word. If a
          table has four words, think a better name.
     ii.  If the column is a primary key, then its name must be xxx_id
     iii. If the column is a date, then its name must be xxx_<action>_dt
          (where <action> represents what's this date use for, e.g.: last
          modification date should be xxx_modification_dt).
    iv.   If the column is a foreign key it must be xxx_<referenced_column>
          where <referenced_column> is the name of the column we want to
          enforce (e.g. task.tas_id would be scheduled_task.sct_tas_id)
          Because of no enforced foreign key control in sqlite we are using
          triggers to keep it error prone.

 SQLConnection: a mini ORM to abstract the commonest queries and handle
                several regular database actions: connect, commit, rollback,
                etc.

"""
# standard library imports
import gettext
import sqlite3 as sql
import cPickle as pic
import types
import os
import os.path
from inspect import getargspec
# related 3rd party imports

# local application/ library specific imports.

# establish translation context
gettext.bindtextdomain('Ghosts', 'locale')
gettext.textdomain('Ghosts')
_ = gettext.gettext

__all__ = ["SQLError", "CommonProperty", "SQLConnection", "createdb"]
_ro_tables = ['temporal_basis', 'watch_type']

class SQLError(Exception):
    """A generic error class for Database Errors"""
    pass

def search_or_create():
    """Returns the path to the database. In case the dir does not exist
    it creates it and then returns it"""
    dir = os.path.expanduser('~/.Ghosts')
    if not os.path.exists(dir):
        os.mkdir(dir)
        createdb()
    return os.path.join(dir, 'main.db')

class CommonProperty(object):
    """Ref. Python CookBook 2d ed., by Alex Martelli, Anna Martelli
    Ravenscroft and David Ascher (O'Reilly Media, 2005) 0-596-00797-3
    Recipe 20.5 by Raymond Hettinger, page 752
    """

    def __init__(self, realname, fget=getattr, fset=setattr, fdel=delattr,
                 doc=None):
        self.realname = realname
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        self.__doc__ = doc or ""
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError(_("Can't get attribute"))
        return self.fget(obj, self.realname)
    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError(_("Can't set attribute"))
        self.fset(obj, self.realname, value)
    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError(_("Can't delete attribute"))
        self.fdel(obj, self.realname)

def adapt_bool(Bool):
    """An adapter for storing boolean values"""
    return int(Bool)

def convert_bool(Bool):
    """A converter for retrieving boolean values"""
    return bool(int(Bool))

def adapt_tuple(Tuple):
    """An adapter for storing tuples"""
    return sql.Binary(pic.dumps(Tuple, pic.HIGHEST_PROTOCOL))

def convert_tuple(Tuple):
    """A converter for retrieving tuples"""
    try:
        return pic.loads(str(Tuple))
    except TypeError:
        if Tuple is None:
            return None
        else:
            raise SQLError(_("""The system was unable to retrieve a \
valid tuple for a given parameter. Value: %s""" % (Tuple)))

def adapt_list(List):
    """An adapter for storing lists"""
    return sql.Binary(pic.dumps(List, pic.HIGHEST_PROTOCOL))

def convert_list(List):
    """A converter for retrieving lists"""
    try:
        return pic.loads(str(List))
    except TypeError:
        if List is None:
            return None
        else:
            raise SQLError(_("""The system was unable to retrieve a \
valid list for a given parameter. Value: %s""" % (List)))

def adapt_dict(Dict):
    """An adapter for storing dicts"""
    return sql.Binary(pic.dumps(Dict, pic.HIGHEST_PROTOCOL))

def convert_dict(Dict):
    """A converter for retrieving dicts"""
    try:
        return pic.loads(str(Dict))
    except TypeError:
        if Dict is None:
            return None
        else:
            raise SQLError(_("""The system was unable to retrieve a \
valid dict for a given parameter. Value: %s""" % (Dict)))

class SQLConnection(object):
    """Base clase for handling connections with the database"""

    _debug = False

    def __init__(self, timeout_=5):
        """Returns connection and cursor objects"""
        sql.register_adapter(bool, adapt_bool)
        sql.register_converter("boolean", convert_bool)
        sql.register_adapter(tuple, adapt_tuple)
        sql.register_converter("tuple", convert_tuple)
        sql.register_adapter(list, adapt_list)
        sql.register_converter("list", convert_list)
        sql.register_adapter(dict, adapt_dict)
        sql.register_converter("dict", convert_dict)
        self.conn = sql.connect(search_or_create(), timeout=timeout_,
        detect_types=sql.PARSE_DECLTYPES)
        self.cursor = self.conn.cursor()

    def disconnects(self):
        """Closes the database connection"""
        self.conn.close()

    def commit(self):
        """Commit the transaction"""
        self.conn.commit()

    def rollback(self):
        """Rollback the transaction"""
        self.conn.rollback()

    def get(self, cls, pk_id, **kwargs):
        """Retrieves information from a table using the primary key. It
        should always retrieve just one row (duh!).
        If no information is found a SQLError will be raised.

        Parameters:
            cls: calling object. The class' object needs a __table__ and a
                 __pk__ attributes stating the table's name and the primary
                key's column(s)' name(s). In case it's a composity a tuple
                should be sent.
            pk_id: the primary key's value we'll use to query.
            kwargs: a "column name":"alias" to use for the query. If no
                    value is provided it will execute a "select * from ...".
                    Otherwise, just the columns listed will be retrieved.
                    Example:
                        kwargs = 'tas_from'='from', 'tas_max'='max'
                        will be turned into
                        select tas_from as from, tas_max as max from ...
        """
        cols = "select "
        if not kwargs:
            cols += "* "
        else:
            for i in kwargs.iterkeys():
                cols += "%s as %s, " % (i, kwargs[i])
        try:
            qry_str = "%s from %s " % (cols.rstrip(', '), cls.__table__)
            if hasattr(cls, '__references__'):
                qry_str += "inner join %s on (%s=%s) " % (cls.__references__,
                        cls.__pk__, cls.__fk__)
            qry_str += " where %s = ?" % (cls.__pk__)
            if self._debug:
                print qry_str
        except AttributeError:
            raise SQLError(_("A needed attribute is missing for this object"))
        try:
            self.cursor.execute(qry_str, (pk_id,))
        except sql.OperationalError, errmsg:
            raise SQLError("%s. String: %s" % (errmsg, qry_str))
        row = self.cursor.fetchone()
        try:
            return dict(zip([x[0] for x in self.cursor.description],
                            (y for y in row)))
        except TypeError:
            raise SQLError(_("Selected value does not exist in table %s" \
                             % (cls.__table__)))

    def update(self, cls, updatedValue, updatedCriteria, commit=True):
        """Updates a table.
           In case updatedValue and updatedCriteria are lists (of dicts)
           will execute a bulk update (à la executemany). If they're dicts
           will execute just one update (see respective parameter docs for
           more on this).
           In case it's a bulk update both updatedCriteria and updatedValue
           must be a list of dicts.

        Parameters:
            cls: calling object. The class' object needs a __table__ and a
                 __pk__ attributes stating the table's name and the primary
                key's column(s)' name(s). In case it's a composity a tuple
                should be sent.
            updatedValue: a "column name":"update value" dict to use for the
                    query.
                    Example:
                        values = {'tas_from'=u'/tmp',
                        'tas_add_date'='2008-03-20 12:34:56.0'}
                        will be turned into
                        update __table__ set tas_from=u'/tmp',
                        tas_add_date='2008-03-20 12:34:56.0'
                        where ...
            updatedCriteria: if its a single value, will match the __pk__
                    attribute for the class. If its a dict it will containe
                    the the column=filter criteria that will be concatenated
                    with "and".
                    Additional tips:
                    1- If you need to specify a more complex criteria
                    (e.g. with "or") use the word 'SQL' as key and
                    write the exact 'where' filter you need. SQL will
                    be append to the rest of the filters with an
                    'and' unless it's the only one filter.
                    2- If you want to use a "like" expression the
                    value will be scannned to find a % token.
                    3- To pass a "is null" criteria, use None as the
                    key's value.
                    4- To pass a "is not null" criteria, use '!None'
                    as the key's value.
                    5- If "filter" is a tuple or list it will be use
                    to construct an "in (tuple)" filter.
                    6- If updatedCriteria is None all table will be
                    updated.
            commit: indicates if the statement should be committed or not.
                    Default Yes
        """
        if cls.__table__ in _ro_tables:
            raise SQLError(_("""The selected table cannot be modified"""))
        cols = "update %s set " % (cls.__table__)
        # check whether we might need to executemany :)
        if type(updatedValue) == list and type(updatedCriteria) == list:
            stm = "executemany"
        else:
            stm = "execute"
        bindval = list()
        qry_str = ""
        # TODO: no time to find out a way to factorize this: might do some day
        if stm == "execute":
            for i in updatedValue.iterkeys():
                cols += "%s=?, " % (i)
                bindval.append(updatedValue[i])
            if type(updatedCriteria) == dict:
                qry_str += "where "
                for i in [x for x in updatedCriteria.iterkeys() if \
                          x != 'SQL']:
                    if updatedCriteria[i] is None:
                        qry_str += "%s is null and " % (i)
                    elif type(updatedCriteria[i]) in (unicode, str) and \
                         updatedCriteria[i].find('%') != -1:
                        qry_str += "%s like ? and " % (i)
                        bindval.append(updatedCriteria[i])
                    elif type(updatedCriteria[i]) in (unicode, str) and \
                         updatedCriteria[i] == '!None':
                        qry_str += "%s is not null and " % (i)
                    elif type(updatedCriteria[i]) in (tuple, list) and \
                         updatedCriteria[i]:
                        aux_str = "?, "
                        aux_str *= len(updatedCriteria[i])
                        qry_str += "%s in (%s) and " % (i,
                                                        aux_str.rstrip(", "))
                        bindval += updatedCriteria[i]
                    else:
                        qry_str += "%s=? and " % (i)
                        bindval.append(updatedCriteria[i])
                if updatedCriteria.has_key('SQL'):
                    #qry_str = qry_str.rstrip("and ")
                    qry_str += updatedCriteria['SQL']
                qry_str = "%s %s" % (cols.rstrip(', '), qry_str)
            elif type(updatedCriteria) in (int, str, unicode):
                bindval.append(updatedCriteria)
                qry_str = "%s where %s=?" % (cols.rstrip(', '), cls.__pk__)
            elif not updatedCriteria:
                qry_str = cols.rstrip(', ')
            else:
                raise SQLError(_("""Unable to detect updatedCriteria \
datatype"""))
        else:
            for c, L1 in enumerate(updatedValue):
                bindval.append([])
                for i in L1.iterkeys():
                    if c == 0:
                        cols += "%s=?, " % (i)
                    bindval[c].append(L1[i])
            for c, L1 in enumerate(updatedCriteria):
                if type(L1) == dict:
                    if c == 0:
                        qry_str = "where "
                    if L1.has_key('SQL'):
                        raise NotImplementedError(_("""Bulk updates do not \
support ad-lib SQL expressions"""))
                    for i in [x for x in L1.iterkeys() if \
                              x != 'SQL']:
                        if L1[i] is None:
                            if c == 0:
                                qry_str += "%s is null and " % (i)
                        elif type(L1[i]) in (unicode, str) and \
                             L1[i].find('%') != -1:
                            if c == 0:
                                qry_str += "%s like ? and " % (i)
                            bindval[c].append(L1[i])
                        elif type(L1[i]) in (unicode, str) and \
                             L1[i] == '!None':
                            if c == 0:
                                qry_str += "%s is not null and " % (i)
                        elif type(L1[i]) in (tuple, list) and \
                             L1[i]:
                            if c == 0:
                                aux_str = "?, "
                                aux_str *= len(L1[i])
                                qry_str += "%s in (%s) and " % (i,
                                                        aux_str.rstrip(", "))
                            bindval[c].append(L1[i])
                            #params += L1[i]
                        else:
                            if c == 0:
                                qry_str += "%s=? and " % (i)
                            bindval[c].append(L1[i])
                    if c == 0:
                        qry_str = "%s %s" % (cols.rstrip(', '), qry_str)
                elif type(L1) in (int, str, unicode):
                    bindval[c].append([L1])
                    if c == 0:
                        qry_str = "%s where %s=?" % (cols.rstrip(', '),
                                                     cls.__pk__)
                elif not L1:
                    raise SQLError(_("""No filter criteria established for a \
bulk update"""))
                else:
                    raise SQLError(_("""Unable to detect updatedCriteria\
 datatype"""))
        if self._debug:
            print qry_str.rstrip(" and")
        # we execute out query here
        getattr(getattr(self, "cursor"), stm) (qry_str.rstrip(" and"),
                                               tuple(bindval))
        if commit:
            self.conn.commit()

    def delete(self, cls, deletedCriteria=None, commit=True):
        """Deletes tuples from a table that match certain criteria.

        Parameters:
            cls: calling object. The class' object needs a __table__ and a
                 __pk__ attributes stating the table's name and the primary
                key's column(s)' name(s). In case it's a composity a tuple
                should be sent.
            deletedCriteria: if its a single value, will match the __pk__
                    attribute for the class. If its a dict it will containe
                    the the column=filter criteria that will be concatenated
                    with "and".
                    Additional tips:
                    1- If you need to specify a more complex criteria
                    (e.g. with "or") use the word 'SQL' as key and
                    write the exact 'where' filter you need. SQL will
                    be append to the rest of the filters with an
                    'and' unless it's the only one filter.
                    2- If you want to use a "like" expression the
                    value will be scannned to find a % token.
                    3- To pass a "is null" criteria, use None as the
                    key's value.
                    4- To pass a "is not null" criteria, use '!None'
                    as the key's value.
                    5- If "filter" is a tuple or list it will be use
                    to construct an "in (tuple)" filter.
                    6- If deletedCriteria is None all table will be
                    deleted.
            commit: indicates if the statement should be committed or not.
                    Default Yes
        """
        if cls.__table__ in _ro_tables:
            raise SQLError(_("""The selected table cannot be modified"""))
        cols = "delete from %s " % (cls.__table__)
        if not deletedCriteria:
            qry_str = cols
        else:
            bindval = list()
            if type(deletedCriteria) == dict:
                qry_str = "where "
                for i in [x for x in deletedCriteria.iterkeys() \
                            if x != 'SQL']:
                    if deletedCriteria[i] is None:
                        qry_str += "%s is null and " % (i)
                    elif type(deletedCriteria[i]) in (unicode, str) and \
                         deletedCriteria[i].find('%') != -1:
                        qry_str += "%s like ? and " % (i)
                        bindval.append(deletedCriteria[i])
                    elif type(deletedCriteria[i]) in (unicode, str) and \
                         deletedCriteria[i] == '!None':
                        qry_str += "%s is not null and " % (i)
                    elif type(deletedCriteria[i]) in (tuple, list) and \
                         len(deletedCriteria[i])>0:
                        aux_str = "?, "
                        aux_str *= len(deletedCriteria[i])
                        qry_str += "%s in (%s) and " % (i,
                                                        aux_str.rstrip(", "))
                        bindval += deletedCriteria[i]
                    else:
                        qry_str += "%s=? and " % (i)
                        bindval.append(deletedCriteria[i])
                if deletedCriteria.has_key('SQL'):
                    qry_str = qry_str.rstrip("and ")
                    qry_str += deletedCriteria['SQL']
                qry_str = "%s %s" % (cols.rstrip(', '), qry_str)
            elif type(deletedCriteria) in (int, str, unicode):
                bindval.append(deletedCriteria)
                qry_str = "%s where %s=?" % (cols.rstrip(', '), cls.__pk__)
            else:
                raise SQLError(_("""Unable to detect deletedCriteria \
datatype"""))
        if self._debug:
            print qry_str.rstrip(" and")
        if not deletedCriteria:
            self.cursor.execute(qry_str)
        else:
            self.cursor.execute(qry_str.rstrip(" and"), tuple(bindval))
        if commit:
            self.conn.commit()

    def insert(self, cls, values, constraint_action, commit=True):
        """Inserts a new value in the database. Returns the primary key value
        in case the queried class defines one.

        Parameters:
            cls: calling object. The class' object needs a __table__ and a
                 __pk__ attributes stating the table's name and the primary
                key's column(s)' name(s). In case it's a composity a tuple
                should be sent.
            values: a "column name":"inserted value" dict to use for the
                    query.
                    Example:
                        values = {'tas_from':u'/tmp',
                        'tas_add_date'='2008-03-20 12:34:56.0'}
                        will be turned into
                        insert into __table__ (tas_from, tas_add_date)
                        values (u'/tmp', '2008-03-20 12:34:56.0')
            if "values" is a list of dicts (following the up supra rules)
            it will perform a bulk execute (à la executemany)
            constraint_action: what action the system must apply in case
                               of contraint violation. The possible values
                               are the ones supported by the SQL engine:
                               ABORT, REPLACE, IGNORE, ROLLBACK, FAIL.
            commit: indicates if the statement should be committed or not.
                    Default Yes
        """
        # TODO: if we have a table with just 1 column and that column is
        #       __pk__ this method fails: fix it!
        if cls.__table__ in _ro_tables:
            raise SQLError(_("""The selected table cannot be modified"""))
        ret_pk = None
        if constraint_action == 'replace':
            cols1 = "replace into %s (" % (cls.__table__)
        elif constraint_action == 'abort':
            cols1 = "insert into %s (" % (cls.__table__)
        else:
            cols1 = "insert or %s into %s (" % (constraint_action.lower(),
                                                cls.__table__)
        cols2 = " values ("
        # check whether we might need to executemany :)
        if type(values) == dict:
            stm = "execute"
        elif type(values) == list:
            stm = "executemany"
        else:
            raise SQLError(_("You need to provide the values to insert"))
        bindval = list()
        # TODO: no time to find out a way to factorize this: might do some day
        if stm == "execute":
            for i in [i for i in values.iterkeys() if \
                      (hasattr(cls, "__pk__") and i != cls.__pk__) or \
                      not hasattr(cls, "__pk__")]:
                cols1 += "%s, " % (i)
                cols2 += "?, "
                bindval.append(values[i])
        else:
            for c, L1 in enumerate(values):
                params = []
                for i in [i for i in L1.iterkeys() if \
                          (hasattr(cls, "__pk__") and i != cls.__pk__) or \
                          not hasattr(cls, "__pk__")]:
                    if c == 0:
                        cols1 += "%s, " % (i)
                        cols2 += "?, "
                    params.append(L1[i])
                bindval.append(params)
        qry_str = "%s) %s) " % (cols1.rstrip(', '), cols2.rstrip(', '))
        if self._debug:
            print qry_str
        # we execute out query here
        getattr(getattr(self, "cursor"), stm) (qry_str, tuple(bindval))
        # if defined, we return the newly pk generated
        if constraint_action != "replace" and hasattr(cls, '__pk__') and \
        stm == "execute":
            self.cursor.execute("""select max(%s) from %s""" % (cls.__pk__,
            cls.__table__))
            ret_pk = self.cursor.fetchone()[0]
        if commit:
            self.conn.commit()
        return ret_pk

    def find(self, cls, selectColumns=None, filterCriteria=None,
             returnAsInstance=False):
        """Find the results in a table given by cls.__table__ and return
        a list of dicts with the object's attributes that fullfill
        filterCriteria.

        Parameters:
            cls: object to query
            selectColumns: the columns the query must return. If it's a
                           tuple it will return these columns as if.
                           If it's a dictionary it will use the key for the
                           column and the value as the column's alias.
                           If None it will execute a "select * ..."
            filterCriteria: a dict containing the the column=filter criteria
                            that will be concatenated with "and".
                            Additional tips:
                            1- If you need to specify a more complex criteria
                            (e.g. with "or") use the word 'SQL' as key and
                            write the exact 'where' filter you need. SQL will
                            be append to the rest of the filters with an
                            'and' unless it's the only one filter.
                            2- If you want to use a "like" expression the
                            value will be scannned to find a % token.
                            3- To pass a "is null" criteria, use None as the
                            key's value.
                            4- To pass a "is not null" criteria, use '!None'
                            as the key's value.
                            5- If "filter" is a tuple or list it will be use
                            to construct an "in (tuple)" filter.
                            6- If filterCriteria is None all table will be
                            retrieved.
            returnAsInstance: in certain cases user will want to retrieve a
                              list of live objects: if this parameter is True
                              will call the cls.__init__ for every tuple
                              returned by the database. Default to False.
        """
        ret_objs = list()
        qry_str = "select "
        if not selectColumns:
            qry_str += "* "
        else:
            if type(selectColumns) is tuple:
                cols = ", ".join(x for x in selectColumns)
                qry_str += cols
            elif type(selectColumns) is dict:
                for i in selectColumns.iterkeys():
                    qry_str += "%s as %s, " % (i, selectColumns[i])
                qry_str = qry_str.rstrip(', ')
        qry_str += " from %s " % (cls.__table__)
        if hasattr(cls, '__references__'):
            qry_str += "inner join %s on (%s=%s) " % (cls.__references__,
                    cls.__pk__, cls.__fk__)
        flt_qry_str = list()
        if filterCriteria:
            qry_str += "where "
            for i in [x for x in filterCriteria.iterkeys() if x != 'SQL']:
                if filterCriteria[i] is None:
                    qry_str += "%s is null and " % (i)
                elif type(filterCriteria[i]) in (unicode, str) and \
                     filterCriteria[i].find('%') != -1:
                    qry_str += "%s like ? and " % (i)
                    flt_qry_str.append(filterCriteria[i])
                elif type(filterCriteria[i]) in (unicode, str) and \
                     filterCriteria[i] == '!None':
                    qry_str += "%s is not null and " % (i)
                elif type(filterCriteria[i]) in (tuple, list) and \
                     filterCriteria[i]:
                    aux_str = "?, "
                    aux_str *= len(filterCriteria[i])
                    qry_str += "%s in (%s) and " % (i, aux_str.rstrip(", "))
                    flt_qry_str += filterCriteria[i]
                else:
                    qry_str += "%s=? and " % (i)
                    flt_qry_str.append(filterCriteria[i])
            if filterCriteria.has_key('SQL'):
                qry_str = qry_str.rstrip("and ")
                qry_str += filterCriteria['SQL']
            qry_str = qry_str.rstrip("and ")
        if hasattr(cls, '__pk__'):
            qry_str += " order by %s" % (cls.__pk__)
        if self._debug:
            print qry_str
        if flt_qry_str:
            self.cursor.execute(qry_str, tuple(flt_qry_str))
        else:
            self.cursor.execute(qry_str)
        rows = self.cursor.fetchall()
        for row in rows:
            final_val = dict(zip([x[0] for x in self.cursor.description],
                                (y for y in row)))
            if returnAsInstance:
                # we assume you really know what's your doing.
                params = getargspec(cls.__init__)
                # do they accept a kwargs?
                if params[2]:
                    ret_objs.append(cls(None, **final_val))
                elif params[0][1] == 'id':
                    ret_objs.append(cls(final_val[cls.__pk__]))
            else:
                ret_objs.append(final_val)
        return ret_objs

def ObjectFactory(cls, ParamDict=None):
    """Returns list of instances (as read only objects) of requested class
    Parameters:
        * cls: selected class to be queried
        * ParamDict: dict with filter criteria to apply in the form
                     attribute:value (see createdb.SQLConnection.find for more
                     on this)
    """
    sqlcon_ = SQLConnection()
    if cls.__name__ not in ('Task', 'TemporalBasis', 'Transport', 'WatchFile'):
        raise NotImplementedError(_("""Sorry, we don't currently support that \
specific class"""))
    #get the data
    # little hack to fix an issue with Transport:
    if cls.__name__ == 'Transport' and (not ParamDict or not
            ParamDict.has_key('sto_cancel_date')):
        ParamDict = dict(sto_cancel_date=None)
    rc = sqlcon_.find(cls, None, ParamDict, True)
    # if class has an __included_cls__ attr. it means some of its attributes
    # might be instantiated too, if that's the case, let's do it.
    if hasattr(cls, '__included_cls__'):
        for x in rc:
            for y in cls.__included_cls__.iteritems():
                setattr(x, y[1].__name__, y[1](getattr(x, y[0])))
    # make all instances read only if this behaviour is accepted.
    if hasattr(x, '_ro'):
        setattr(x, '_ro', True)
    return rc

def createdb(haltOnErrors=True, alwaysDrop=True):
    """Calling this function will create the database.
    Parameters:
        haltOnErrors: boolean. If True, whenever we hit an error the
        def will inform this and stop. If False, it will inform the error
        but keep on with the next statement. This is can be useful if you
        try to create a db that already exists but you might want to add
        missing tables. This flag applies to errors in tables's creation.
        It can't do nothing about missing permissions, e.g. trying to create
        the db in a forbidden place, etc
        alwaysDrop: boolean. If True, in case of '...already exists' error,
        the function will drop the object and retry to create. If False,
        whatever is set in HaltOnError will happen.
"""
    try:
        sql.register_adapter(bool, adapt_bool)
        sql.register_converter("boolean", convert_bool)
        sql.register_adapter(tuple, adapt_tuple)
        sql.register_converter("tuple", convert_tuple)
        sql.register_adapter(list, adapt_list)
        sql.register_converter("list", convert_list)
        sql.register_adapter(dict, adapt_dict)
        sql.register_converter("dict", convert_dict)
        con = sql.Connection(search_or_create())
    except sql.OperationalError, errmsg:
        raise SQLError(_("""There's a problem with the directory where you \
want to store the database. Check permissions or fix the path's name. Error \
message: %s""") % (errmsg))

    # the scripts:
    script = list()

    # the protocol table:
    script.append("""create table protocol (
        pro_id integer primary key,
        pro_name text not null,
        pro_mask dict,
        unique (pro_name)
        )""")

    # Dictionary with the mandatory information we need to gather for each
    # protocol. It holds the attribute as key and three possible values:
    #   M-  Mandatory: this information is required
    #   A-  Alternative: not mandatory but optional
    #   None- This information is not neccessary for the given protocol
    d_ssh = {'path':None, 'user':'A', 'pass':'A', 'ipaddress':'M',
             'port':'M', 'uuid':None, 'private_key':'A'}
    d_ftp = {'path':None, 'user':'A', 'pass':'A', 'ipaddress':'M',
             'port':'M', 'uuid':None, 'private_key':'A'}
    d_usb = {'path':None, 'user':None, 'pass':None, 'ipaddress':None,
             'port':None, 'uuid':'M', 'private_key':None}
    d_mde = {'path':'M', 'user':None, 'pass':None, 'ipaddress':None,
             'port':None, 'uuid':'A', 'private_key':None}
    #script.append("""insert into protocol (pro_name, pro_mask) values \
#('ssh', '%s')""" % (d_ssh))
    #script.append("""insert into protocol (pro_name, pro_mask) values \
#('sftp', '%s')""" % (d_ftp))
    #script.append("""insert into protocol (pro_name, pro_mask) values \
#('usb', '%s')""" % (d_usb))
    #script.append("""insert into protocol (pro_name, pro_mask) values \
#('mdevice', '%s')""" % (d_mde))

    # the temporal_basis table:
    script.append("""create table temporal_basis (
        teb_id integer primary key,
        teb_name text not null,
        teb_validate boolean not null,
        teb_priority integer not null,
        unique (teb_name)
        )""")
    script.append("""create index teb_pk_idx on temporal_basis (teb_id)""")

    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('MONTHLY', 1, 0)""")
    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('WEEKLY', 1, 0)""")
    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('DAILY', 1, 0)""")
    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('HOURLY', 1, 0)""")
    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('On demand', 0, -2)""")
    script.append("""insert into temporal_basis (teb_name, teb_validate,
    teb_priority) values ('In the moment', 0, -1)""")

    # the global_parameter table:
    script.append("""create table global_parameter (
        glp_id text not null,
        glp_value list,
        glp_allow_nonunique bool,
        glp_allow_delete bool,
        unique (glp_id)
        )""")

    # the storage table:
    script.append("""create table storage (
        sto_id integer primary key,
        sto_name text not null,
        sto_base text not null,
        sto_transport text not null,
        sto_add_date timestamp not null,
        sto_cancel_date timestamp,
        unique (sto_name)
        )""")
    script.append("""create index sto_pk_idx on storage (sto_id)""")
    script.append("""create index sto_base_idx on storage (sto_base)""")
    script.append("""create index sto_name_idx on storage (sto_name)""")

    # the task table:
    script.append("""create table task (
        tas_id integer primary key,
        tas_sto_id integer not null constraint tas_sto_fk references
                   storage(sto_id),
        tas_from text not null,
        tas_to text,
        tas_max tuple,
        tas_teb_id integer not null constraint tas_teb_fk references
                   temporal_basis(teb_id),
        tas_time_coord dict,
        tas_including list,
        tas_excluding list,
        tas_tas_id integer constraint tas_tas_fk references task(tas_id),
        tas_add_date timestamp not null,
        tas_cancel_date timestamp,
        tas_last_update integer
        )""")

    script.append("""create trigger tas_bef_insert_trg
        before insert on task
        for each row begin
            select raise(rollback, 'Task: Foreign key violation, temporal basis
            does not exist in the origin table')
            where new.tas_teb_id is not null and
            (select teb_id from temporal_basis where teb_id=NEW.tas_teb_id)
            is null;
            select raise(rollback, 'Task: Foreign key violation, task id
            does not exist in the origin table')
            where new.tas_tas_id is not null and
            (select tas_id from task where tas_id=NEW.tas_tas_id)
            is null;
            select raise(rollback, 'Task: Foreign key violation, storage id
            does not exist in the origin table')
            where new.tas_sto_id is not null and
            (select sto_id from storage where sto_id=NEW.tas_sto_id)
            is null;
        end;""")
    script.append("""create trigger tas_aft_insert_trg
        after insert on task
        begin
            update task set tas_last_update = strftime('%s', 'now')
            where rowid = last_insert_rowid();
        end;""")
    script.append("""create trigger tas_aft_update_trg
        after update on task
        begin
            update task set tas_last_update = strftime('%s', 'now')
            where rowid = last_insert_rowid();
        end;""")
    script.append("""create index tas_pk_idx on task (tas_id)""")
    script.append("""create index tas_teb_idx on task (tas_teb_id)""")
    script.append("""create index tas_last_idx on task (tas_last_update)""")

    # the task schedule table
    script.append("""create table task_schedule (
        tac_id integer primary key,
        tac_tas_id integer not null constraint tac_tas_fk references
                   task(tas_id),
        tac_planned_date timestamp,
        tac_execution_date timestamp,
        tac_elapsed integer default 0,
        tac_success boolean,
        tac_log text
        )""")
    script.append("""create index tac_tas_idx on task_schedule
        (tac_tas_id)""")
    script.append("""create trigger tac_insert_trg
        before insert on task_schedule
        for each row begin
            select raise(rollback, 'Task Schedule: Foreign key violation, task
            id does not exist in the origin table')
            where new.tac_tas_id is not null and
            (select tas_id from task where tas_id=NEW.tac_tas_id)
            is null;
        end;""")

    # the watch type table:
    script.append("""create table watch_type (
        wat_id integer primary key,
        wat_name text
    )""")
    script.append("""insert into watch_type (wat_name) values ('Included')""")
    script.append("""insert into watch_type (wat_name) values ('Excluded')""")
    script.append("""insert into watch_type (wat_name) values \
    ('Cancelled')""")

    # the watched files table:
    script.append("""create table watched_file (
        waf_id integer primary key,
        waf_tas_id integer not null constraint waf_tas_fk references
                   task(tas_id),
        waf_wat_id integer not null constraint waf_wat_fk references
                   watch_type(wat_id),
        waf_name text integer not null,
        waf_path text integer not null,
        waf_size integer not null,
        waf_sha text,
        waf_owner text,
        waf_perms text,
        waf_last_known_status text,
        waf_last_date timestamp,
        unique (waf_tas_id, waf_name, waf_path)
        )""")
    script.append("""create index waf_tas_idx on watched_file
        (waf_tas_id)""")
    script.append("""create index waf_wat_idx on watched_file
        (waf_wat_id)""")
    script.append("""create trigger waf_insert_trg
        before insert on watched_file
        for each row begin
            select raise(rollback, 'Watched File: Foreign key violation, task id
            does not exist in the origin table')
            where new.waf_tas_id is not null and
            (select tas_id from task where tas_id=NEW.waf_tas_id)
            is null;
            select raise(rollback, 'Watched File: Foreign key violation, watch
            type does not exist in the origin table')
            where new.waf_wat_id is not null and
            (select wat_id from watch_type where wat_id=NEW.waf_wat_id)
            is null;
        end;""")

    # the temp watch descriptor table
    script.append("""create table inotify_descriptor (
        ind_tas_id integer not null,
        ind_path text not null,
        ind_wd_id integer,
        ind_timestamp timestamp,
        ind_flag integer,
        unique (ind_tas_id, ind_path)
        )""")
    script.append("""create index ind_wd_idx on inotify_descriptor
        (ind_wd_id)""")
    script.append("""create index ind_path_idx on inotify_descriptor
        (ind_path)""")
    script.append("""create index ind_flag_idx on inotify_descriptor
        (ind_flag)""")

    # the events table
    script.append("""create table event (
        eve_seq integer primary key,
        eve_wd integer not null,
        eve_tas_id integer not null,
        eve_path text,
        eve_name text,
        eve_pathname text,
        eve_mask int,
        eve_maskname text,
        eve_dir bool,
        eve_cookie integer,
        eve_timestamp timestamp
        )""")
    script.append("""create index eve_tas_idx on event (eve_tas_id)""")

    # the coordinator_special_event table
    script.append("""create table coordinator_special_event (
        cse_action text not null,
        cse_timestamp timestamp not null
        )""")

    # the job table:
    script.append("""create table job (
        job_id integer primary key,
        job_tas_id integer not null,
        job_from text not null,
        job_including list,
        job_excluding list,
        job_sto_id integer not null,
        job_to text not null,
        job_priority integer not null,
        job_scheduled_date timestamp not null,
        job_timestamp integer not null
        )""")
    script.append("""create index job_idx_pk on job (job_id)""")
    script.append("""create index job_tas_idx on job (job_tas_id)""")
    script.append("""create index job_tms_idx on job (job_timestamp)""")
    script.append("""create index job_tasdt_idx on
    job(job_tas_id,job_scheduled_date)""")

    script.append("""create table job_detail (
        jod_id integer primary key,
        jod_job_id integer not null,
        jod_path text not null,
        jod_file text not null,
        jod_waf_id integer not null,
        jod_timestamp integer not null,
        jod_backup_date timestamp,
        jod_backup_name text,
        jod_log text
        )""")
    script.append("""create index jod_idx_pk on job_detail (jod_id)""")
    script.append("""create index jod_job_idx on job_detail (jod_job_id)""")
    script.append("""create index jod_search_idx on job_detail (jod_path,
    jod_file, jod_timestamp)""")

    # VIEWS
    script.append("""create view v$to_schedule_task as select tas_id, teb_name
    tas_teb_id, tas_time_coord from task inner join temporal_basis on
    (tas_teb_id=teb_id) where tas_cancel_date is null and teb_validate=1""")
    script.append("""create view v$event_set as select eve_seq, eve_tas_id,
    eve_path, eve_name from event""")

    # the main loop
    for val in script:
        try:
            con.execute(val)
        except StandardError, errmsg:
            if not haltOnErrors:
                print errmsg
                pass
            else:
                if errmsg[0].endswith("already exists"):
                    if alwaysDrop:
                        if errmsg[0].split()[1].startswith("v$"):
                            con.execute("drop view %s" % (
                                        errmsg[0].split()[1]))
                        else:
                            con.execute("drop %s %s" % (errmsg[0].split()[0],
                            errmsg[0].split()[1]))
                        con.execute(val)
                    else:
                        raise SQLError(_("""There's a problem with a \
database instruction. Fix the problem and retry. \nStatement: %s \nError \
message: %s""") % (val, errmsg))
                else:
                    raise SQLError(_("""There's a problem with a \
database instruction. Fix the problem and retry. \nStatement: %s \nError \
message: %s""") % (val, errmsg))
        except OperationalError:
            print val
        #finally:
        #    con.close()
    l1 = [u'/tmp']
    l2 = [u'swp', u'swpx', u'~', u'swx', u'pyc']
    con.execute("""insert into global_parameter (glp_id, glp_value, \
glp_allow_nonunique, glp_allow_delete) values ('global_excluded_extension', \
?, ?, ?)""", (l2, True, True))
    con.execute("""insert into global_parameter (glp_id, glp_value, \
glp_allow_nonunique, glp_allow_delete) values ('global_temporary_storage', \
?, ?, ?)""", (l1, False, False))
    con.execute("""insert into global_parameter (glp_id, glp_value, \
glp_allow_nonunique, glp_allow_delete) values ('global_maximum_thread', \
?, ?, ?)""", ([3], False, False))
    con.execute("""insert into global_parameter (glp_id, glp_value, \
glp_allow_nonunique, glp_allow_delete) values ('global_sleep_interval', \
?, ?, ?)""", ([60], False, False))
    con.execute("""insert into global_parameter (glp_id, glp_value, \
glp_allow_nonunique, glp_allow_delete) values ('global_backup_method', \
?, ?, ?)""", (['rdiff-backup'], False, False))

    con.commit()
