# Copyright (C) 2008 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import task as t

# +++ Creating a task +++

# a Task instance can be generated in 2 different ways:

# === 1 ===
# if it does not exist: create it

# the __init__ will accept a dictionary with the required attributes. To find
# out the valid attributes you need to read the class help, specifically the
# 'Data descriptors...' section
# help(t.Task)

# I list them here:
"""
>>> [x for x in dir(t.Task) if x.startswith('tas_')]
['tas_add_date', 'tas_cancel_date', 'tas_from', 'tas_excluding', 'tas_id',
'tas_including', 'tas_max', 'tas_sto_id', 'tas_tas_id', 'tas_teb_id',
'tas_time_coord', 'tas_to']
"""

# Of all of them,
# * tas_add_date, tas_cancel_date and tas_tas_id will be ignored
#   if you set them since they're used internally.
# * tas_id is a primary key and upon task creation will be set by the database,
#   if you pass it as an attribute it means you want to update and existing task
# * regarding tas_teb_id: this value maps with task.TemporalBasis class: you
#   can pass an id (mapping an existing one in the database) or an instance of
#   TemporalBasis
"""
>>> tb = t.TemporalBasis(1)
>>> tb
<task.TemporalBasis object at 0x8e2a1cc>
>>> tb.teb_id, tb.teb_name, tb.teb_validate
(1, u'MONTHLY', True)
"""
# * tas_sto_id: this value maps with task.Transport class: you can pass an id
# (mapping an existing and valid one in the database) or an instance of
# Transport (see examples/Transport.py for more on this class)
"""
>>> tr = t.Transport(2)
>>> tr
<task.Transport object at 0x8e2bb8c>
>>> tr.sto_id, tr.sto_transport, tr.sto_base
(2, u'bzr+ssh://', u'mmara@10.23.1.2/home/mmara')
"""
# For the rest of the attributes, please read the data descriptors where you
# will find -I hope- enough info to build them

# let's first create a dict with all the required values and fill it up with
# valid data:
new_dict = dict((x,None) for x in dir(t.Task) if x.startswith('tas_'))
new_dict.pop('tas_id')
new_dict.pop('tas_add_date')
new_dict.pop('tas_cancel_date')
new_dict.pop('tas_tas_id')
"""
>>> new_dict
{'tas_sto_id': None, 'tas_max': None, 'tas_including': None, 'tas_to': None,
'tas_excluding': None, 'tas_time_coord': None, 'tas_teb_id': None,
'tas_from': None}
"""
new_dict['tas_including'] = u'*.py'
new_dict['tas_excluding'] = u'*.aspx'
# tas_teb_id: we're passing an id although we could use a temporal basis
# instance too (if you use a temporal basis instance it doesn't matter whether
# the key is named tas_teb_id or anything else since the __init__ checks if
# it's an instance of the base class or not)
new_dict['tas_teb_id'] = 2
# tas_sto_id: we're passing an id although we could use a Transport instance
# too or the base value (If you use a Transport instance it doesn't matter
# whether the key is named tas_sto_id or anything else since __init__ checks if
# it's an instance of the base class or not)
new_dict['tas_sto_id'] = 2
new_dict['tas_to'] = u'/mybackup'
new_dict['tas_from'] = u'/home/mariano/Music'
# you should read the tas_time_coord __doc__ string to fully understand how it
# works.
new_dict['tas_time_coord'] = {'DW':[1,4,5,6]}
"""
>>> new_dict
{'tas_from': u'/home/mariano/Music', 'tas_sto_id': 2, 'tas_max': None,
'tas_including': u'*.py', 'tas_to': u'/mybackup', 'tas_excluding': u'*.aspx',
'tas_time_coord': {'DW': [1, 4, 5, 6]}, 'tas_teb_id': 2}
"""
# let's call the class
ta = t.Task(None, **new_dict)
# As you can see, we pass a None value as first
# parameter since we can use it to retrieve an existing task which, of course,
# is not this case
"""
>>> ta.tas_id, ta.tas_from, ta.tas_including, ta.tas_excluding, ta.tas_sto_id,
ta.tas_time_coord, ta.tas_add_date
(None, u'/home/mariano/Music', [u'*.py'], [u'*.aspx'], 2, {'DW': [1, 4, 5, 6]},
datetime.datetime(2008, 6, 20, 3, 39, 40, 900318))
"""
# The task id is still null: we need to commit the changes to assign one
# (although we're not gonna do it right now)
# let's see what happens if we modify the tas_time_coord attribute, adding bad
# values and then invoke method "validate_time_coord"
"""
>>> ta.tas_time_coord = {'DW':[1,76,5555,4,7,7]}
>>> ta.validate_time_coord()
>>> ta.tas_time_coord
{'DW': [1, 4], 'HD': [datetime.time(0, 0)]}
"""
# as you can see we added invalid values (the Temporal Basis was defined as
# Weekly so only values between 0 and 6 are allowed). When you validate, all
# invalid members are deleted.

# let's commit the changes:
ta.commit()
"""
>>> ta.tas_id, ta.tas_from, ta.tas_including, ta.tas_excluding, ta.tas_sto_id,
ta.tas_time_coord, ta.tas_add_date
(1, u'/home/mariano/Music', [u'*.py'], [u'*.aspx'], 2, {'DW': [1, 4, 5, 6]},
datetime.datetime(2008, 6, 20, 3, 39, 40, 900318))
"""

# A few additional comments:
# * tas_including and tas_excluding are lists: you can add as many values as
# you want: you can use absolute paths, complete files names, supported glob
# idioms and regular expresions (to use regexp, you need to preappend the
# expression with RE:, otherwhise it will not work).
# only Temporal Basis Monthly, Weekly, Daily and Hourly require a
# tas_time_coord: the other two will always ignore any value stored for them.

# === 2 ===
# you can init a task using the task identification number (pass it as a first
# argument to init, of course it should exist in the database)
del ta
ta = t.Task(1)
"""
>>> [x for x in ta.__dict__.iteritems() if x[0].startswith('tas_')]
[('tas_max', (None,)), ('tas_sto_id', 2), ('tas_cancel_date', None), ('tas_to',
u'/mybackup'), ('tas_excluding', [u'*.aspx']), ('tas_id', 1), ('tas_from',
u'/home/mariano/Music'), ('tas_add_date', datetime.datetime(2008, 6, 20, 3, 39,
40, 900318)), ('tas_tas_id', None), ('tas_last_update', 1213944586),
('tas_time_coord', {'DW': [1, 4], 'HD': [datetime.time(0, 0)]}),
('tas_including', [u'*.py']), ('tas_teb_id', 2)]
"""

# as you can see all previously stored values are retrieved.

# if you try to get an unexisting task identification number, you'll raise an
# error:
"""
>>> ta = t.Task(1300)
Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
        File "task.py", line 606, in __init__
            rc = sqlcon.get(self, id)
              File "createdb.py", line 234, in get
                  % (cls.__table__)))
createdb.SQLError: Selected value does not exist in table task
"""

# +++ Cancelling a task +++

# Now it's time to try a cancellation:
del ta
ta = t.Task(1)
ta.cancel()
# it's done: let's check it:
"""
>>> ta.tas_cancel_date
datetime.datetime(2008, 6, 23, 1, 30, 39, 173263)
"""
# if we try to update any of the class's attributes we get an error:
"""
>>> ta.tas_teb_id = 4
Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
        File "createdb.py", line 102, in __set__
            self.fset(obj, self.realname, value)
              File "task.py", line 690, in _set
                  cannot be modified"))
task.ReadOnlyError: The task is cancelled and cannot be modified
"""

# +++ Modifying a task +++

# after you instanced an active task you can change whatever attribute you want
# and save it again: there will exist a few changes you need to know:

# * the modified task will be cancelled and a new task will be automatically
# generated inheriting all the unchanged attributes from its parent plus the
# ones you changed. In order to track which task is a modified one you have an
# atribute tas_tas_id. If this attr. is not null it will carry the id of the
# parent task.

# Let's see an example
del ta
ta = t.Task(4)
# now we modify an attribute. As you can notice, there are no methods for this:
# you set the attribute directly (plain ol' python):
ta.tas_teb_id = 4
# save it:
ta.commit()
# we can see the tas_id changed:
"""
>>> ta.tas_id
15
>>> ta.tas_tas_id
4
"""
# If you instance task 5 again you'll see it's cancelled and it cannot be
# modify anymore. All watched files will be transferred to the new task in the
# moment of the creation of the task
# IMPORTANT:
# Python does not allow tracking changes in lists, tuples and dicts made
# with their own methods (e.g. pop(), append()) so the Task is unable to
# realize a change in attributes tas_max, tas_including, tas_excluding and
# tas_time_coord unless the attribute is directly modified. Task relies on
# noticing the attribute changed to efectively call update method().
# You can fix this situation with two methods (that I can think right now):
# * update _update[atrribute] to true. e.g ta._update['tas_including'] = True
# * modify another attribute (even putting back the same value if had) so that
#   the _update dictionary gets updated and the commit can proceed.

# +++ Getting all tasks +++

# You can get a list of all Tasks (each of them as instance) calling the
# ObjectFactory method.
import createdb as c
rc = c.ObjectFactory(t.Task, None)
"""
>>> print rc
[<task.Task object at 0x84f13ec>, <task.Task object at 0x85e20ac>, <task.Task
object at 0x84f88ec>, <task.Task object at 0x85e224c>, <task.Task object at
0x85e236c>, <task.Task object at 0x85e242c>, <task.Task object at 0x85e24ec>,
<task.Task object at 0x85e25ac>, <task.Task object at 0x85e266c>, <task.Task
object at 0x85e272c>, <task.Task object at 0x85e286c>, <task.Task object at
0x85e294c>, <task.Task object at 0x85e2a2c>, <task.Task object at 0x85e2b0c>,
<task.Task object at 0x85e2bec>, <task.Task object at 0x85e2ccc>, <task.Task
object at 0x85e2dac>, <task.Task object at 0x85e2e8c>, <task.Task object at
0x85e2f6c>, <task.Task object at 0x85e20ec>, <task.Task object at 0x85e210c>,
<task.Task object at 0x84f1c8c>]
"""
# a few notes on this method:
# Parameters:
#    cls = The class you want to retrieve [1]
#    criteria: you can pass a dict with filter criteria if you want (an example
#              right ahead) [2]
#   [1]: currently it only supports Task only.
#   [2]: see an example below =>
"""
>>> # bring all tasks that has a Transport id 1
...
>>> rc = c.ObjectFactory(t.Task, {'tas_sto_id':1})
>>> print rc
[<task.Task object at 0x8d6b3cc>, <task.Task object at 0x8d6bbec>, <task.Task
object at 0x8d6b56c>, <task.Task object at 0x8d6bd4c>, <task.Task object at
0x8d6be2c>, <task.Task object at 0x8d6beec>, <task.Task object at 0x8d6bfac>,
<task.Task object at 0x8d6bc2c>, <task.Task object at 0x8d6bc4c>, <task.Task
object at 0x8d6bc0c>, <task.Task object at 0x8d6e32c>, <task.Task object at
0x8d6e40c>]
"""
#   Additional notes:
#   * All objects in the list will be created readonly and cannot be modified
#   or save back to the database: if you want to modify one of them, you should
#   update _ro private attribute back to False. Note that the save() "might" fail
#   anyway. Might be a better idea to instance a new task using the tas_id of
#   the selected object.
#   * Two additional attributes are created for each object: a Transport and a
#   TemporalBasis: these two are related to tas_sto_id and tas_teb_id
#   respectively. This two attributes might save some time in order not to look
#   up in the database the meaning of the tas_sto_id and/or tas_teb_id. Worth
#   noticing that both Transport and TemporalBasis are instances of the values
#   found in the normal attributes of Task. See an example below =>
"""
>>> rc[0].Transport
<task.Transport object at 0x8d6b42c>
>>> rc[0].Transport.sto_id, rc[0].tas_sto_id
(1, 1)
>>> rc[0].Transport.sto_base
u'/home/mariano/Downloads'
>>> rc[0].Transport.sto_transport
u'file://'
>>> rc[0].TemporalBasis.teb_id, rc[0].TemporalBasis.teb_name, rc[0].tas_teb_id,
(6, u'In the moment', 6)
"""
# If you want a lighter method to get a/some given task(s) -e.g. just the tasks
# data with a smaller footprint- you can access directly the mini-ORM API.
# Here's an example
"""
>>> import createdb as c
>>> sqlcon = c.SQLConnection()
>>> import task as t
>>> rc = sqlcon.find(t.Task, None, None, False)
>>> len(rc)
22
>>> rc[0]
{'tas_cancel_date': datetime.datetime(2008, 6, 23, 1, 30, 39, 173263),
'tas_from': u'/home/mariano/Music', 'tas_max': (None, None), 'tas_sto_id': 2,
'tas_add_date': datetime.datetime(2008, 6, 19, 16, 39, 13, 411446), 'tas_to':
u'test', 'tas_excluding': [u'*.m4a'], 'tas_id': 1, 'tas_including': [u'*.mp3'],
'tas_tas_id': None, 'tas_last_update': 1213904463, 'tas_time_coord': {'DW': [1,
6], 'HD': [datetime.time(0, 0)]}, 'tas_teb_id': 2}
"""
# This method will return a list (one element for each Task) of dicts (the key
# is the attribute's name). The last parameter (True | False) tells the method
# if you want data or instances (that's why we are passing False in this case)
# This method allows you to just ask for the values you're interested in:
"""
>>> rc = sqlcon.find(t.Task, ('tas_id', 'tas_from', 'tas_including'), None,
>>> False)
>>> rc[0]
{'tas_including': [u'*.mp3'], 'tas_id': 1, 'tas_from': u'/home/mariano/Music'}
"""
# As you can see, you pass a tuple as the second parameter, listing all
# attributes you want to retrieve.
# The third parameter allows you to filter the tasks you want to get back, see
# an example:
"""
>>> rc = sqlcon.find(t.Task, None, dict(tas_sto_id=1), False)
>>> rc[0]
{'tas_cancel_date': datetime.datetime(2008, 6, 24, 14, 26, 14, 555866),
'tas_from': u'/home/mariano/sourcecode/Ghosts/trash', 'tas_max': (None, None),
'tas_sto_id': 1, 'tas_add_date': datetime.datetime(2008, 6, 20, 17, 35, 47,
21725), 'tas_to': u'testing', 'tas_excluding': [u'*.txt', u'*.jpg'], 'tas_id':
4, 'tas_including': [u'*.py'], 'tas_tas_id': None, 'tas_last_update':
1213994151, 'tas_time_coord': None, 'tas_teb_id': 6}
"""
# The drawback of this method is that you need to run another query to get, for
# example, the details of the assigned Transport:
"""
>>> tr = sqlcon.get(t.Transport, rc[0]['tas_sto_id'])
>>> tr
{'sto_base': u'/home/mariano/Downloads', 'sto_add_date':
datetime.datetime(2008, 6, 19, 15, 49, 47, 106407), 'sto_transport':
u'file://', 'sto_id': 1, 'sto_cancel_date': None, 'sto_name': u'Downloads'}
"""

