# Copyright (C) 2008 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""This represents all C.R.U.D. actions the user needs to apply in order to
make the program work.

The main class is Task. It represents a job in the system scheduled to
perform a specific backup task. All others classes are somehow related to
Task and are defined to provide additional support to schedule a Task and
help it fulfill its duties.

"""

# standard library imports
import gettext
import datetime as dt
import calendar as cal
import itertools
import os
import stat
import types
import time

# local application/ library specific imports.
import createdb as cdb
from bzrlib import globbing, osutils, transport, errors
from watcher import InotifyDescriptor

# connects with the database
sqlcon = cdb.SQLConnection()
# establish translation context
gettext.bindtextdomain('Ghosts', 'locale')
gettext.textdomain('Ghosts')
_ = gettext.gettext

class TaskError(Exception):
    """Base class for all errors in this module"""
    pass

class GlobalParameterError(TaskError):
    """Error handling class for errors derived from Global Parameter"""
    pass

class TaskScheduleError(TaskError):
    """Error handling class for errors derived from Task Schedule"""
    pass

class ReadOnlyError(TaskError):
    """This error will be raised when user try to redefine an attribute that
    for the given state of the object is forbidden"""
    pass

class WatchedFileError(TaskError):
    """Error handling class for errors derived from Watch File"""
    pass

class TransportError(TaskError):
    """Error handling class for errors related to transport"""
    pass

class GlobalParameter(object):
    """This class defines attributes that apply to every task.

    global_excluded_extension = list of extensions that must be globally
    excluded from all backup jobs. Typically this means temporal files
    (such as .swp files generated by vim or compiled bytecode files such
    as Python's pyc). If user adds an extension here it will be globally
    ignored and never backup. User should only submit the extension with
    no glob parameter and no dot

    global_temporary_storage = in case all destination backup devices can't be
    accessed (e.g. network down on a ssh transport or unmounted device in
    a file transport), the backup job will be stored in this temporary storage
    until the system can safely backup in the desired transport. If no one is
    defined by user, will default to /tmp

    global_maximum_thread = how many concurrent threads may exist to perform
    backup services

    global_sleep_interval = how much time can the backup daemon sleep before
    checking if a new scheduled taks exists (in seconds)

    global_backup_method = the name of the strategy/algorithm selected to
    perform the job.

    This class defines three types of methods:
    * a boilerplate set_xx where xx represents the attribute name as described
    up supra (e.g. set_global_temporary_storage) and a simple parameter that
    indicates the value to be set. This method will commit so it should not be
    called as part of an ongoing transaction.
    * a boilerplate get_xx where xx represents the attributed name as
    described up supra (e.g. get_global_temporary_storage) that takes no
    parameter and will return the set parameter(s).
    * a boilerplate del_xx where xx represents the attributed name as
    described up supra (e.g. del_global_temporary_storage) that takes a
    simple parameter with the value to be erased.

    """

    __table__ = "global_parameter"
    __pk__ = "glp_id"

    def _make_func(self, value):
        """Build methods required to get and set global parameters"""
        rc = self.gpsqlcon.find(self, ("glp_value",), dict(glp_id=value))
        def make_get():
            rc = self.gpsqlcon.find(self, ("glp_value",), dict(glp_id=value))
            return rc[0]['glp_value']
        make_get.__name__ = "get_%s" % (value.encode("ascii"))
        def make_set(new_value):
            rc = self.gpsqlcon.find(self, ("glp_value", "glp_allow_nonunique"),
                        dict(glp_id=value))
            new_list = rc[0]['glp_value']
            if not rc[0]['glp_allow_nonunique']:
                new_list[0] = new_value
            else:
                new_list.append(new_value)
            self.gpsqlcon.update(self, dict(glp_value=new_list), value, True)
        make_set.__name__ = "set_%s" % (value.encode("ascii"))
        def make_del(old_value):
            rc = self.gpsqlcon.find(self, ("glp_value", "glp_allow_delete"),
                                    dict(glp_id=value))
            if not rc[0]['glp_allow_delete']:
                raise GlobalParameterError(_("""This value cannot be \
deleted"""))
            else:
                rc[0]['glp_value'].remove(old_value)
                self.gpsqlcon.update(self, dict(glp_value=rc[0]['glp_value']),
                                     value, True)
        make_del.__name__ = "del_%s" % (value.encode("ascii"))
        return (make_get, make_set, make_del)

    def __init__(self, sqlcon_=None):
        if sqlcon_:
            self.gpsqlcon = sqlcon_
        else:
            self.gpsqlcon = sqlcon
        rc = self.gpsqlcon.find(self, ("glp_id",))
        for a in rc:
            (f1, f2, f3) = self._make_func(a['glp_id'])
            setattr(self, "%s" % (f1.__name__), f1)
            setattr(self, "%s" % (f2.__name__), f2)
            setattr(self, "%s" % (f3.__name__), f3)

class TemporalBasis(object):
    """Represents the possible temp options for scheduling

    The possible values are: Monthly (the user can select the desired days
    when to backup), Weekly (the user can select the day(s) when she/he wants
    to backup), Daily (he user can select the hour(s) when to backup), Hourly
    (the user can select the minute(s) when to backup), On Demand (the user
    will specify the moment when she/he want to backup) and In the moment
    (selected files will be watched and as soon there's a change in them,
    they'll be backuped).

    Attributes:
    teb_id = unique identification
    teb_name = name of the selected temporal basis
    teb_validate = boolean value that indicates if the selected temporal
    basis should be checked upon creation

    """

    __table__ = "temporal_basis"
    __pk__ = "teb_id"

    class TemporalBasisError(Exception):
        """Error handling class for errors derived from Temporal Basis"""
        pass

    def __init__(self, id=None):
        """The id parameter will retrieve the selected temporal
        basis. If None, will create an attribute for temporal basis but
        user will need to init again with the selected teb
        """
        if id:
            # we connect with the database to get a record
            rc = sqlcon.get(self, id)
            for x in rc.items():
                setattr(self, x[0], x[1])
        rc = sqlcon.find(self, ('teb_id', 'teb_name'))
        for cur in rc:
            setattr(self, cur['teb_name'].replace(' ', '').upper(),
                    cur['teb_id'])

class WatchFile(object):
    """This class is meant to handle anything related to files on disk
    The main actions this class need to handle are:
        1- find the files the user wants to backup: walk the dir tree,
           apply the selected filters (Globster) and save the data in the
           database.
        2- retrieve the selected files from database and compare them to
           the actual tree and find out which ones have changed (check the
           size and the sha).
        3- Walk the tree and find new files added to the tree that match the
           task's filters.
    """

    __table__ = "watched_file"
    __pk__ = "waf_id"

    def __init__(self, id=None, sqlcon_=None):
        """Get a list of to-watch files.
        If id is provided get the info from the database, otherwise initialize
        all attributes to empty lists

        Parameters:
            id = identification number of the task.
            sqlcon_ = In case of need, ignore the module's database connection
            and use on submitted by the calling party
        """
        # we define here the macros for the watch types.
        if sqlcon_:
            self.SQLCon = sqlcon_
        else:
            self.SQLCon = sqlcon
        rc = self.SQLCon.cursor.execute("""select * from watch_type""")
        for row in self.SQLCon.cursor.fetchall():
            setattr(self, row[1].lower(), row[0])
        del rc

        self.tas_id = id

        if self.tas_id:
            self.load_watched_files()
        else:
            self._init_hook()

    def load_watched_files(self):
        """"Load previous state of all files watched from the database"""
        # now find the watched files since we got a task id.
        if self.tas_id:
            rc = self.SQLCon.find(self, None, dict(waf_tas_id=self.tas_id))
            if rc:
                for cur1 in rc:
                    for cur2 in list(cur1.iteritems()):
                        if hasattr(self, cur2[0]):
                                   getattr(getattr(self, cur2[0]), "append") \
                                  (cur2[1])
                        else:
                            setattr(self, cur2[0], [cur2[1]])
            else:
                self._init_hook()

    def _init_hook(self):
        """Call in to create a add a few attributes in case they cannot be
        retrieved from the datatabase"""
        self.waf_tas_id = list()
        self.waf_last_date = list()
        self.waf_last_known_status = list()
        self.waf_name = list()
        self.waf_path = list()
        self.waf_wat_id = list()
        self.waf_id = list()
        self.waf_size = list()
        self.waf_sha = list()
        self.waf_owner = list()
        self.waf_perms = list()

    def get_changed_files(self, callback=None):
        """Compare a list of saved files with their real counterparts and find
        out the ones that have changed. A callback function can be defined and
        it will be executed before ending every iteration
        """
        updated_values = []
        updated_criteria = []

        for idx, val in enumerate(self.waf_id):
            if self.waf_wat_id[idx] == self.included:
                
                new_file_name = os.path.join(self.waf_path[idx],
                                             self.waf_name[idx])

                new_size = os.path.getsize(new_file_name)
                new_owner = os.stat(new_file_name)[stat.ST_UID]
                new_perms = oct(os.stat(new_file_name)[stat.ST_MODE])

                if new_size != self.waf_size[idx]:

                    try:
                        new_sha = osutils.sha_file(file(new_file_name,
                                                   'rb', buffering=65000))
                    except IOError, errmsg:
                        self.SQLCon.update(self,
                        {"waf_last_date":dt.datetime.now(),
                        "waf_last_known_status":errmsg.args[1]},
                        {"waf_id":self.waf_id[idx]}, True)
                        continue
                    updated_values.append({"waf_size":new_size,
                    "waf_sha":new_sha, "waf_last_date":dt.datetime.now(),
                    "waf_owner":new_owner, "waf_perms":new_perms,
                    "waf_last_known_status":"File size changed"})
                    updated_criteria.append({"waf_id":self.waf_id[idx],
                    "waf_name":self.waf_name[idx],
                    "waf_path":self.waf_path[idx]})

                else:
                    try:
                        new_sha = osutils.sha_file(file(new_file_name,
                                                   'rb', buffering=65000))
                    except IOError, errmsg:
                        self.SQLCon.update(self,
                        {"waf_last_date":dt.datetime.now(),
                        "waf_last_known_status":errmsg.args[1]},
                        {"waf_id":self.waf_id[idx]}, True)
                        continue
                    if new_sha != self.waf_sha[idx]:
                        updated_values.append({"waf_size":new_size,
                        "waf_sha":new_sha, "waf_last_date":dt.datetime.now(),
                        "waf_owner":new_owner, "waf_perms":waf_perms,
                        "waf_last_known_status":"File size changed"})
                        updated_criteria.append({"waf_id":self.waf_id[idx],
                        "waf_name":self.waf_name[idx],
                        "waf_path":self.waf_path[idx]})

            self.SQLCon.update(self, updated_values, updated_criteria, True)
            if callback:
                callback(self)

        # Reload from the database
        for l in [x for x in self.__dict__.keys() if x.startswith('waf_')]:
            del self.__dict__[l][:]
        self.load_watched_files()

        return updated_criteria

    def watch_file(self, from_, tas_id=None):
        """The directory specified in from_ will be mark to be watched by
        inotify (including subdirs by default).
        No including, excluding filters will be applied right now.
        Parameters:
            from_: See Task.tas_from doc string
            tas_id: task identification integer
        """
        if not self.waf_tas_id and tas_id:
            self.waf_tas_id = tas_id
        to_watch = InotifyDescriptor(self.SQLCon)
        to_watch.insert(self.waf_tas_id, from_)

    def stop_watch(self, tas_id):
        """Will erase a watch from the watcher's list"""
        if not self.waf_tas_id and tas_id:
            self.waf_tas_id = tas_id
        to_watch = InotifyDescriptor(self.SQLCon)
        to_watch.delete(self.waf_tas_id)

    def track_file(self, from_, including=[], excluding=[],
                   find_new=False, tas_id=None, return_values=False):
        """The directory specified in from_, including and
        excluding should be scanned and register for later retrieval.
        Returns two lists in case parameter return_values is True: included
        and excluded
        Parameters:
            from_: See Task.tas_from doc string
            including: see Task.tas_including doc string
            excluding: see Task.tas_excluding doc string
            find_new: if True will search for new files and ignore any
                      files previously loaded. This applies to the database
                      only. All related work will be performed anyway
            tas_id: task identification integer
            return_values: if True will return two lists (included and
                           excluded)
        """
        # TODO: should find of way of not walking the whole tree again to
        # find new files.
        if find_new:
            sel_action = 'ignore'
            # we build this dict to make already present file look up faster.
            # As key we have (path, file) and the watch flag as value
            fast_search = dict(zip(zip(self.waf_path, self.waf_name),
                self.waf_wat_id))
        else:
            sel_action = 'replace'
        included = list()
        excluded = list()
        if not self.waf_tas_id and tas_id:
            self.waf_tas_id = tas_id
        # we build two globsters with including and excluding
        # if including is empty, we assume everything
        if not including:
            including.append(u'*')
        # here we add global excluded extensions
        gp = GlobalParameter(self.SQLCon)
        excluded += map(lambda s: "*." + s,
                        gp.get_global_excluded_extension())
        glob_inc = globbing.Globster(including)
        glob_exc = globbing.Globster(excluding)
        # for every file in from_, let's check the files to see if they
        # match including, excluding, both or neither.
        fl_l = list(os.walk(from_))
        # TODO: it's possible that user wants to in/exclude a directory: not
        # current support for that selection yet.
        for idx1, idx2, idx3 in fl_l:
            for idx4 in idx3:
                # it's included but excluded too and the last one is in charge
                # or
                # excluded and not included
                if (glob_inc.match(idx4) and glob_exc.match(idx4)) or\
                    (not glob_inc.match(idx4) and glob_exc.match(idx4)):
                    if find_new:
                        val = fast_search.get((idx1, idx4))
                        # if the already present value does not match the watch
                        # type we need to update the db.
                        if val and val != self.excluded:
                            self.SQLCon.update(self,
                                {"waf_wat_id":self.excluded, "waf_size":0, 
                                "waf_sha":"0", "waf_owner":0, "waf_perms":0,
                                "waf_last_known_status":"Exclusion update",
                                "waf_last_date":dt.datetime.now()},
                                {"waf_tas_id":self.waf_tas_id, "waf_name":idx4,
                                "waf_path":idx1}, False)
                        else:
                            # nothing changed, keep on walking
                            continue
                    try:
                        # in excluded, we might not calculate sha and size
                        # since we don't really need them and in case tomorrow
                        # their status changes we should start from fresh to
                        # backup them no matter if they change or not.
                        self.SQLCon.insert(self, {"waf_tas_id":self.waf_tas_id,
                        "waf_wat_id":self.excluded, "waf_name":idx4,
                        "waf_path":idx1, "waf_size":0, "waf_sha":"0",
                        "waf_owner":0, "waf_perms":0,
                        "waf_last_date":dt.datetime.now()}, sel_action, False)
                    except IOError, errmsg:
                        self.SQLCon.insert(self, {"waf_tas_id":self.waf_tas_id,
                        "waf_wat_id":self.cancelled, "waf_name":idx4,
                        "waf_path":idx1, "waf_size":0, "waf_sha":"0",
                        "waf_owner":0, "waf_perms":0,
                        "waf_last_known_status":errmsg.args[1],
                        "waf_last_date":dt.datetime.now()},
                        sel_action, False)
                    if return_values:
                        excluded.append(idx4, idx1)
                # included and not excluded
                elif glob_inc.match(idx4) and not glob_exc.match(idx4):
                    if find_new:
                        val = fast_search.get((idx1, idx4))
                        # if the already present value does not match the watch
                        # type we need to update the db.
                        if val and val != self.included:
                            self.SQLCon.update(self,
                                {"waf_wat_id":self.included,
                                "waf_last_known_status":"Inclusion update",
                                "waf_last_date":dt.datetime.now(),
                                "waf_owner":0, "waf_perms":0,
                                "waf_size":0, "waf_sha":"0"},
                                {"waf_tas_id":self.waf_tas_id, "waf_name":idx4,
                                "waf_path":idx1}, False)
                        else:
                            # nothing changed, keep on walking
                            continue
                    try:
                        self.SQLCon.insert(self, {"waf_tas_id":self.waf_tas_id,
                        "waf_wat_id":self.included, "waf_name":idx4,
                        "waf_path":idx1, "waf_size":0, "waf_sha":"0",
                        "waf_owner":0, "waf_perms":0,
                        "waf_last_date":dt.datetime.now()}, sel_action, False)
                    except IOError, errmsg:
                        self.SQLCon.insert(self, {"waf_tas_id":self.waf_tas_id,
                        "waf_wat_id":self.cancelled, "waf_name":idx4,
                        "waf_path":idx1, "waf_size":0, "waf_sha":"0",
                        "waf_owner":0, "waf_perms":0,
                        "waf_last_known_status":errmsg.args[1],
                        "waf_last_date":dt.datetime.now()},
                        sel_action, False)
                    if return_values:
                        included.append(idx4, idx1)
        if return_values:
            return included, excluded

class Transport(object):
    """This class encapsulates all supported transports

    Users just need to create a transport for the root of the system where they
    plan to store backups since they can provide additional paths when they
    create a task. e.g.: a valid transport base could be /media/disk and then
    add /backups/dir when they create a task (so the system will look up
    /media/disk/backups/dir. This design is useful in order to create just the
    required transport that later on can be refined selecting specific dirs
    where to backup
    """

    __table__ = "storage"
    __pk__ = "sto_id"
    _new_transport = True
    _user = None
    _password = None
    _host = None
    _port = None
    _relpath = None
    transport = None
    sto_transport = None
    backup_ark_ident_file = '.backupark_validate'

    def __init__(self, id=None, **kwargs):
        """If id is not null we'll get the defined transport in the
        database
        * If id is null, we'll use base to create a new transport, using the
          "name" parameter to personalize it in the database.
        * If both are null a TranportError will be raised
        * Valid formats for "base" are:
                a path to a directory (e.g. /home/user)
                a ssh connection string in the format
                user@host[:port]/home/user. If port is not provided will
                default to 22
        * If base is a remote site, user need to provide a password.
        * In case we can get the transport as a valid database entry the
          appropiate transport will be returned, otherwise user will need
          to execute validate to get the transport
        """
        if not id and not kwargs['sto_base']:
            raise TransportError(_('You have to select either a task id '
                                   'or provide a valid base'))
        else:
            if id:
                rc = sqlcon.get(self, id)
                for x in rc.items():
                    setattr(self, x[0], x[1])
                if self.sto_cancel_date:
                    raise TransportError(_('The selected transport is '
                                           'cancelled'))
                self._new_transport = False
                self.transport = transport.get_transport("%s%s" %
                (self.sto_transport, self.sto_base))
            elif kwargs:
                rc = sqlcon.find(self, None, dict(sto_base=kwargs['sto_base']))
                if rc:
                    for x in rc[0].items():
                        setattr(self, x[0], x[1])
                    if self.sto_cancel_date:
                        raise TransportError(_('The selected transport is '
                                               'cancelled'))
                    self._new_transport = False
                    self.transport = transport.get_transport("%s%s" %
                    (self.sto_transport, self.sto_base))
                else:
                    self.sto_base = kwargs['sto_base']
                    try:
                        self._password = kwargs['password_']
                    except KeyError:
                        pass
                    self.sto_id = None
                    self.sto_name = kwargs['sto_name']
                    self.sto_transport = kwargs['sto_transport']

    def validate(self):
        """Verify the location can be accessed and writen to.
        """

        if os.path.isfile(self.sto_base):
            raise TransportError(_("""Base cannot be a file"""))
        else:
            self.transport = transport.get_transport(self.sto_transport + \
                                                     self.sto_base)
            try:
                self.transport.put_bytes(self.backup_ark_ident_file,
                                         '')
            except errors.NoSuchFile:
                raise TransportError(_('Cannot write in remote directory'))

    def save(self):
        """Save the new transport in the database for later retrieval
        External process calling this method is responsible for committing
        changes in the database: it can use sqlcon.commit() as globally defined
        in this module"""
        if self._new_transport:
            self.sto_id = sqlcon.insert(self, dict(sto_base=self.sto_base,
                          sto_transport=self.sto_transport,
                          sto_name=self.sto_name,
                          sto_add_date=dt.datetime.now()), 'abort', False)

    def cancel(self):
        """Eliminate an existing transport from the database
        External process calling this method is responsible for committing
        changes in the database: it can use sqlcon.commit() as globally defined
        in this module"""
        self.sto_cancel_date = dt.datetime.now()
        if not self._new_transport:
            sqlcon.update(self, {'sto_cancel_date':self.sto_cancel_date},
            {'sto_id':self.sto_id}, False)

class Task(object):
    """This class represents a backup job.

    This class is invoked when there's a need for a new task or
    to update an existing one.

    All updates on the Task database must be executed through this class.

    """

    # table's name and primary key: needed to update the database
    __table__ = "task"
    __pk__ = "tas_id"
    __included_cls__ = {'tas_sto_id':Transport, 'tas_teb_id':TemporalBasis}
    # does the selected value exists in the database?
    _exists = False
    # boolean flag indicating if the task is read only
    _ro = False
    # we store the passed transport in __init__ in case we need to perform
    # additional validations
    _transport = None
    # not provided by user, might be needed in certain scenarios
    tas_id = None
    # a temporal basis instance placeholder to perform certain specific
    # controls.
    _tb = None
    # was tas_time_coord validated?
    _tas_time_coord = False

    def __init__(self, id=None, **kwargs):
        """Creates a new Task unless an id is provided.

        In case there's a task id it will set the information from the
        database.
        If no Task is found in the database with the given id, a
        SQLError will be raised

        Parameters:
        id: the existing identification integer for a task.
        kwargs: key=value entries for all the attributes in the class

        Some words about the **k param:
            * you can find the valid keys to use in this init, checking the
            data descriptors for this class.
            * if you pass tas_id as one of the parameters it will assume the
            task exists in the database and will try to update it with the new
            info. If you're sure the task exists, don't load it with all the *k
            arguments, just pass the "id" parameter with the task id and leave
            kwargs null
            * Of all of them there are a few that you don't need to pass and
            if you do, they'll be ignored or replaced in any moment:
            tas_add_date (will default to datetime.now()), tas_cancel_date
            (use cancel() method or update this attribute directly),
            tas_tas_id (used internally by the system)
            * Of all key=value you can pass, there are two that you can pass
            instances: tas_teb_id (using TemporalBasis) and tas_sto_id (using
            Transport). If you decide to pass instances the words you use to
            define the key doesn't matter since the init will check the instance
            fp.
            * If you don't want to pass instances to parameters tas_teb_id and
            tas_sto_id, you can pass:
                for tas_teb_id, an integer representing the temporal basis id
                selected.
                for tas_sto_id you can pass:
                        1- a sto_id (an integer) in case the storage already
                        exists.
                        2- a sto_base string representing the desired
                        transport: e.g.: user@localhost/home/user for remote
                        servers. The init will try to find the selected base
                        in the db and use it in case it can be found, else
                        it will try to validate it and insert it in the
                        database. If it fails, the init will fail.
            * tas_time_coord is a whole beast on its own. You should read its
            data descriptor doc string to build it succesfully (although a
            "validate" method is provided to verify its integrity before
            loading the info in the database. Actually using validate() is a
            good thing since it provides an *a param to add some time
            definition's shortcuts.
        """

        self._update = dict()
        # we get an id so we get the value from the database
        if id:
            rc = sqlcon.get(self, id)
            for x in rc.items():
                setattr(self, x[0], x[1])
                if x[0] == 'tas_teb_id':
                    self._tb = TemporalBasis(x[1])
            # we find the value in the database, this will lead to an update
            self._exists = True
        # user is giving us the task values
        else:
            for it in (val for val in kwargs.items()): # if val[0] != 'tas_id'):
                if it[0] == 'tas_id' and it[1]:
                    self._exists = True
                if isinstance(it[1], TemporalBasis):
                    setattr(self, 'tas_teb_id', it[1].teb_id)
                    self._tb = TemporalBasis(it[1].teb_id)
                elif isinstance(it[1], Transport):
                    # TODO: check whether we carry on when we cannot check
                    # transport.
                    self._transport = it[1] # save it in case everything fails
                    if not it[1].sto_id:
                        rc = sqlcon.find(Transport, ('sto_id', ),
                        dict(sto_base=it[1].sto_base, sto_cancel_date=None))
                        try:
                            new_sto_id = rc[0]['sto_id']
                        except (TypeError, KeyError): # transp. does not exist
                            new_sto_id = None
                    else:
                        new_sto_id = it[1].sto_id
                    setattr(self, 'tas_sto_id', new_sto_id)
                else:
                    # check the transport:
                    if it[0] == 'tas_sto_id':
                        if type(it[1]) == int:
                            rc = sqlcon.find(Transport, ('sto_id', ),
                            dict(sto_id=it[1], sto_cancel_date=None))
                        else:
                            rc = sqlcon.find(Transport, ('sto_id', ),
                            dict(sto_base=it[1], sto_cancel_date=None))
                        try:
                            setattr(self, it[0], rc[0]['sto_id'])
                        except (IndexError, TypeError, KeyError):
                            # transp. does not exist
                            setattr(self, it[0], None)
                    # check the temporal basis
                    elif it[0] == 'tas_teb_id' and it[1]:
                        self._tb = TemporalBasis(it[1])
                        setattr(self, it[0], it[1])
                    # check datatypes
                    elif it[0] == 'tas_max' and type(it[1]) != tuple:
                        setattr(self, it[0], tuple([it[1]]))
                    elif it[0] == 'tas_time_coord' and \
                         type(it[1]) not in  (dict, types.NoneType):
                        setattr(self, it[0], dict(it[1]))
                    elif it[0] in ('tas_including', 'tas_excluding') and \
                         type(it[1]) != list:
                        setattr(self, it[0], [it[1]])
                    else:
                        setattr(self, it[0], it[1])
            if not hasattr(self, "tas_add_date") or self.tas_add_date is None:
                setattr(self, "tas_add_date", dt.datetime.now())
                self._update['tas_add_date'] = False
            if not hasattr(self, "tas_max"):
                setattr(self, "tas_max", (None, None))
            if not hasattr(self, "tas_including"):
                setattr(self, "tas_including", [])
            if not hasattr(self, "tas_excluding"):
                setattr(self, "tas_excluding", [])

    # following three functions are technically useless since we changed how
    # we track updates, although an _update flag is useful to know when we
    # should call self.update() or self.insert()
    def _get(self, name):
        """Descriptor fget: just perform the usual stuff"""
        return self.__dict__[name]

    def _set(self, name, val):
        """Descriptor fset: every time an attribute is set, record in an
        auxiliary dict _update. First time (for example when using cdb.get)
        mark as False, second time as True"""
        # the cancel date: if set != None, the task is dead and cannot
        # be recreate
        if name == "tas_cancel_date":
            if val is not None:
                self._ro = True
                self.__dict__[name] = val
            else:
                self._ro = False
        if not self._update.has_key(name):
            self._update[name] = False
        else:
            if self._ro is True:
                raise ReadOnlyError(_('The task is cancelled or created as '
                                      'read only and cannot be modified'))
            else:
                self._update[name] = True
        # FIXME: lists and tuples are ignored by the descriptor so we fail to
        # notify changed in tas_max, tas_including and tas_excluding.
        if hasattr(self, name):
            if type(val) in (list, tuple) and val != self.__dict__[name]:
                self._update[name] = True
        self.__dict__[name] = val

    def _del(self, name):
        """Descriptor fdel: just perform the usual stuff"""
        self.__dict__.pop(name)
        self._update.pop(name)

    tas_sto_id = cdb.CommonProperty("tas_sto_id", fget=_get, fset=_set,
                  fdel=_del, doc="""The selected transport""")

    tas_from = cdb.CommonProperty("tas_from", fget=_get, fset=_set,
               fdel=_del, doc="""The directory to backup.""")

    tas_to = cdb.CommonProperty("tas_to", fget=_get, fset=_set,
               fdel=_del, doc="""The directory where to backup. It must be
               a relative path to the selected transport""")

    tas_max = cdb.CommonProperty("tas_max", fget=_get, fset=_set,
            fdel=_del, doc="""A tuple that represents the maximum time the
            user wants to keep the backup before losing changes. The first
            value indicates in minutes the max. time. The second value
            indicates in bytes the max. size of the backup for this
            given folder (the two values can be set: in that case the
            first to happen is the one we use). None represents that
            the value has not been set. If none of them has been set
            then we backup ad infinitum.""")

    tas_teb_id = cdb.CommonProperty("tas_teb_id", fget=_get, fset=_set,
                 fdel=_del, doc="""See TemporalBasis.teb_id""")

    tas_time_coord = cdb.CommonProperty("tas_time_coord", fget=_get,
                     fset=_set, fdel=_del,
            doc="""a dict with the selected periods when to backup
            The type of the data depends on TempBas, e.g.: if the selected
            Temporal Basis is Weekly it will contain a list of days of the
            week mapped to key 'DW' and a list of datetime.time for each day
            mapped to key 'HD', etc.
            The possible keys for this dict are:
                'DM': day of month (works for Monthly). Possible values are
                      1 to 31 (depending on the month). Wrong values (e.g.
                      day 32) will be ignored (this applies to all other keys)
                'DW': day of week (works for Weekly). Possible values are 0
                      to 6 where 0 is Monday (although we recommend using
                      calendar's constant to define weekdays)
                'HD': hour of the day (works for Monthly, Weekly, Daily,
                      Hourly). Possible values are every hour:minute (it must
                      be a datetime.time instance) of the day. In case of
                      Hourly users only can set the minute part.
                'LD': boolean flag (default False) that indicates that the
                      backup should be performed every last day of the month.
                      Only applies when Monthly basis has been selected.
                "On demand" and "In this Moment" do not use this dict.""")

    tas_including = cdb.CommonProperty("tas_including", fget=_get,
                    fset=_set, fdel=_del, doc="""List of all files, subdirs, \
etc we want to include.
            It accepts regexp too (to code a regexp, be sure to add RE:
            before it) and standard linux glob expressions. If this value is
            not provided it will asume everything minus 'excluding'""")

    tas_excluding = cdb.CommonProperty("tas_excluding", fget=_get,
                    fset=_set, fdel=_del, doc="""A list of all files,
                    subdirs, etc to be excluded. It accepts regexp too
                    (see tas_including doc string) for details.""")

    tas_tas_id = cdb.CommonProperty("tas_tas_id", fget=_get, fset=_set,
                 fdel=_del, doc="""In case this task inherited from a prior
                 one, the value of its father""")

    tas_add_date = cdb.CommonProperty("tas_add_date", fget=_get, fset=_set,
                   fdel=_del, doc="The creation date for this task")

    tas_cancel_date = cdb.CommonProperty("tas_cancel_date", fget=_get,
                      fset=_set, fdel=_del, doc="""The cancelation date for \
this task""")

    def cancel(self):
        """The selected task will be cancelled and all events removed"""
        try:
            self.tas_cancel_date = dt.datetime.now()
        except ReadOnlyError:
            pass

        # Erase watched files from inotify daemon and erase tracked files
        # from WatchedFiles (only one of them apply)
        try:
            wf = WatchFile(self.tas_id)
            if self.tas_teb_id != self._tb.INTHEMOMENT:
                sqlcon.update(wf, {'waf_last_date':self.tas_cancel_date,
                'waf_wat_id':wf.cancelled}, dict(waf_tas_id=self.tas_id), False)
            else:
                wf.stop_watch(self.tas_id)

            # Cancel the task.
            sqlcon.update(self, {'tas_cancel_date':self.tas_cancel_date,
                'tas_last_update':int(time.time())}, self.tas_id, False)
            sqlcon.commit()
        finally:
            sqlcon.rollback()

    def commit(self):
        """The task will be saved in the database"""
        # TODO: add validate_time_coord in case self._tas_time_coord is False
        # first check if we have a valid transport:
        if not self.tas_sto_id and not self._ro and not self._exists:
            if isinstance(self._transport, Transport):
                # we were unable to verify the transport up to this point so
                # we have to do it right now or die
                self._transport.validate()
                self._transport.save()
                self.tas_sto_id = self._transport.sto_id
            else:
                raise TaskError(_("""You did not provided a valid transport \
to create a backup destination"""))
        # the value was selected from the database: in case there are changes
        # we will use an update
        if self._exists:
            # the task was cancelled, update related tables and cancel_date
            if self._ro is True:
                self._cancel()
            else:
                # the task already exists, it's not read only and one value was
                # updated: this leads us to an update
                if True in (x for x in self._update.itervalues()):
                    self._update_()
        # the value is a fresh new. We insert this value.
        else:
            self._insert()
        # save event
        sqlcon.commit()

    def validate_time_coord(self, *args):
        """In case of need, validate the time coordinates
        Optionally add a few hints in parameter args: possible values are
        seven_days, first_day and last_day

        """
        # monthly
        if self.tas_teb_id == self._tb.MONTHLY:
            try:
                self.tas_time_coord['DM'] = [val for val in \
                                            self.tas_time_coord['DM']
                                            if val >= 1 and val <= 31]
            except TypeError:
                # this means no DM was defined, nevermind passing this
                # error, we'll catch it later.
                pass
            if 'seven_days' in args:
                self._seven_days()
            if 'first_day' in args:
                self._first_day()
            if 'last_day' in args:
                self._last_day()
            # we got here and no DM has been defined. Unless the last_day
            # shortcut exists, we have to raise an error.
            if (not self.tas_time_coord.has_key('DM') or \
                not self.tas_time_coord['DM']) and 'last_day' not in args:
                raise TaskScheduleError(_("""You have to provide at \
least one valid day to create a Monthly Backup Schedule"""))
            elif (not self.tas_time_coord.has_key('DM') or \
                not self.tas_time_coord['DM']) and 'last_day' in args:
                pass
            else:
                self.tas_time_coord['DM'] = sorted(list(set(
                                            self.tas_time_coord['DM'])))
        # weekly
        elif self.tas_teb_id == self._tb.WEEKLY:
            try:
                self.tas_time_coord['DW'] = sorted(list(set([val for \
                                            val in self.tas_time_coord['DW']
                                              if val >= cal.MONDAY and \
                                                 val <= cal.SUNDAY])))
            except KeyError:
                # here we trap self.tas_time_coord
                raise TaskScheduleError(_("""You have to provide at \
least one valid day to create a Weekly Backup Schedule"""))
                # here we raise if Weekly was selected but no days were
                # provided.
            if not self.tas_time_coord['DW']:
                raise TaskScheduleError(_("""You have to provide at \
least one valid day to create a Weekly Backup Schedule"""))
        # if temporal basis mut be validated, get in here to verify
        # the hours parameter: we filter out not valid time instances.
        if self._tb.teb_validate:
            try:
                self.tas_time_coord['HD'] = sorted(list(set(
                                        itertools.ifilter(
                                        lambda s: isinstance(s, dt.time),
                                                self.tas_time_coord['HD']
                                        ))))
                if not self.tas_time_coord['HD']:
                    if 'last_day' in args and self.tas_teb_id == 1:
                        pass
                    else:
                        self.tas_time_coord['HD'] = [dt.time(0,0,0)]
            except (KeyError, TypeError):
                if 'last_day' in args and self.tas_teb_id == 1:
                    pass
                else:
                    self.tas_time_coord['HD'] = [dt.time(0,0,0)]
        # if it's hourly we remove the hour part since we only care 'bout
        # minutes and seconds
        if self.tas_teb_id == self._tb.HOURLY:
            self.tas_time_coord['HD'] = sorted(list(set(d.replace(hour=0) \
                                        for d in self.tas_time_coord['HD'])
                                        ))
        self._tas_time_coord = True

    def excluded_file(self, file_):
        """Check file and verify if it's excluded"""
        gp = GlobalParameter()
        var_excluding = self.tas_excluding
        var_excluding += map(lambda s: "*." + s,
                        gp.get_global_excluded_extension())
        glob_exc = globbing.Globster(var_excluding)
        del var_excluding
        if glob_exc.match(file_):
            return True
        else:
            return False

    def included_file(self, file_):
        """Check file and verify if it's included"""
        if not self.excluded_file(file_):
            if not self.tas_including:
                self.tas_including = ['*']
            glob_inc = globbing.Globster(self.tas_including)
            if glob_inc.match(file_):
                return True
            else:
                return False
        else:
            return False

    def cluded_file(self, file_):
        """Check file and verify if it's included (True), Excluded (False) or
        None (None)"""
        if self.excluded_file(file_):
            return False
        else:
            if self.included_file(file_):
                return True
            else:
                return None

    def _cancel(self):
        self.cancel()

    def _insert(self):
        # first, let's check if the path exists:
        if not os.path.exists(self.tas_from):
            raise TaskError(_("""The selected path does not exist"""))
        # insert the task
        self.tas_id = sqlcon.insert(self, dict(x for x in \
                                    self.__dict__.iteritems() \
                                    if not x[0].startswith('_')),
                                    'abort', False)
        # update watched files
        if self.tas_teb_id not in (self._tb.INTHEMOMENT,
                                   self._tb.ONDEMAND):
            wf = WatchFile()
            wf.track_file(self.tas_from, self.tas_including,
            self.tas_excluding, False, self.tas_id, False)
        else:
            # since inotify triggers its own commit event, we
            # should be sure the task gets in the table before
            # adding it to the proper watch event
            self._watch_hook()

    def _update_(self):
        # we insert a new task:
        self.tas_tas_id = self.tas_id
        self.tas_id = sqlcon.insert(self, dict(x for x in \
                                    self.__dict__.iteritems() \
                                    if not x[0].startswith('_') and \
                                    x[0] != 'tas_id'),
                                    'abort', False)
        # we cancel the active task: the trick for "updating" is not
        # changing the current values but instead to cancel the task and
        # creating a new one, joining them with tas_tas_id
        sqlcon.update(self, {'tas_cancel_date':dt.datetime.now(),
        'tas_last_update':int(time.time())}, self.tas_tas_id, False)
        wf = WatchFile(self.tas_tas_id)
        # find which was the former temporal basis value:
        former_teb = sqlcon.get(self, self.tas_tas_id,
                **{'tas_teb_id':'tas_teb_id'})['tas_teb_id']
        # we need to cancel old task watched files
        if former_teb != self._tb.INTHEMOMENT:
            sqlcon.update(wf, {'waf_last_date':dt.datetime.now(),
            'waf_wat_id':wf.cancelled}, dict(waf_tas_id=self.tas_tas_id),
            False)
        else:
            wf.stop_watch(self.tas_tas_id)

        # update watched files
        # TODO: in case tas_from didn't change we need to copy the files
        # directly from the last task but not recalculate them since we might
        # lose pending changes.
        wf = WatchFile(self.tas_id)
        if self.tas_teb_id not in (self._tb.INTHEMOMENT,
                                   self._tb.ONDEMAND):
            wf.track_file(self.tas_from, self.tas_including,
            self.tas_excluding, False, self.tas_id, False)
        else:
            self._watch_hook()

    def _seven_days(self):
        """In case a Monthly basis has been selected, this will populate 'DM'
        key with a day from every week of the month: 7, 14, 21, 28"""
        try:
            self.tas_time_coord['DM'] += [7, 14, 21, 28]
        except KeyError:
            self.tas_time_coord['DM'] = [7, 14, 21, 28]

    def _first_day(self):
        """In case a Monthly basis has been selected, this will populate 'DM'
        key with the first day of the month"""
        try:
            self.tas_time_coord['DM'].append(1)
        except KeyError:
            self.tas_time_coord['DM'] = [1]

    def _last_day(self):
        """In case a Monthly basis has been selected, this will activate the
        boolean flag 'LD' (lastDay) in the tas_time_coord dict"""
        self.tas_time_coord['LD'] = True

    def _watch_hook(self):
        """Mark the task's files to be watched by inotify"""
        wf = WatchFile(self.tas_id)
        wf.watch_file(self.tas_from, self.tas_id)
        sqlcon.commit()


    def __eq__(self, other):
        """Will return a list of all the atributes that are equal to attributes
        present in 'other'"""
        if other.__class__ != self.__class__:
            raise TaskError(_("""Can only compare two instances of the same \
class"""))
        rlist = []
        for cur1 in dir(self):
            if cur1.startswith("tas_") and \
                    getattr(self, cur1) == getattr(other, cur1):
                rlist.append(cur1)
        return rlist

    def __ne__(self, other):
        """Will return a list of all the attributes that are not equal to
        attributes present in 'other'"""
        if other.__class__ != self.__class__:
            raise TaskError(_("""Can only compare two instances of the same \
class"""))
        rlist = set([x for x in dir(self) if x.startswith("tas_")])
        return list(rlist.difference(self.__eq__(other)))

    #def __repr__(self):
    #    return "%s %s %s %s %s %s %s %s" % (self.tas_id, self.tas_from,
    #            self.tas_teb_id, self.tas_time_coord, self.tas_including,
    #            self.tas_excluding, self.tas_add_date, self.tas_cancel_date)
