#! /usr/bin/env python2.5
# Copyright (C) 2008 Martin Albisetti 

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""Testing task.TaskSchedule"""

# standard library imports
import unittest
import sys
import sqlite3 as sql
import datetime as dt
# related 3rd party imports
import coverage
from dateutil import relativedelta
# local application/ library specific imports.
sys.path.insert(1,sys.path[0].rsplit('/',1)[0])
import task as t 
import createdb as db

sqlcon = db.SQLConnection()

class testTaskSchedule(unittest.TestCase):

    @staticmethod    
    def compareLists(list1, list2):
        """Compares two list to verify if they share the same values"""
        if len(set(list1).difference(list2)) == 0:
            return True
        else:
            return False

    def testTasExists(self):
        """We load an instance providing an existing tas_id"""
        # first we create a task in the database. We are just providing
        # the not null fields
        vnow = dt.datetime.now()
        sqlcon.cursor.execute("""insert into task (tas_from, tas_teb_id, 
        tas_add_date) values (?, ?, ?)""", ('/tmp', 1, vnow))
        sqlcon.conn.commit()
        sqlcon.cursor.execute("""select tas_id from task where 
        tas_add_date=?""", (vnow,))
        row = sqlcon.cursor.fetchone()
        # checking
        newinst = t.TaskSchedule(row[0])
        self.failUnlessEqual(isinstance(newinst, t.TaskSchedule), True, """We\
 selected a valid tas_id but fail to create instance""")
        
    def testTasNoExists(self):
        """We load an instance providing a non existing tas_id"""
        # we find the last tas_id and then search for an inexistance tas
        sqlcon.cursor.execute("""select max(tas_id) from task""")
        row = sqlcon.cursor.fetchone()
        try:
            v_val = row[0]+1
        except TypeError:
            v_val = 1
        self.failUnlessRaises(db.SQLError, t.TaskSchedule, v_val) 

    def testCreateOk(self):
        """We create a new TaskSchedule object with all ok params"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, "seven_days", DM=[1,15], 
        HD=[dt.time(23,00)])
        self.failUnlessEqual(isinstance(new_ts, t.TaskSchedule), True, """We \
created an instance of TaskSchedule with valid input but it failed""")
    
    def testCreateWrong(self):
        """We create a new TaskSchedule object with an invalid TemporalBasis
        object (meaning a non TemporalBasis object)"""
        class whatever(object):
            dummy1 = None
            dummy2 = None
        new_tb = whatever()
        self.failUnlessRaises(AttributeError, t.TaskSchedule, None, new_tb)

    def testCreateMonthly1(self):
        """We create a new TaskSchedule object with TB == Monthly and just
        one date and no hour"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], [1]), 
        True, """We created a monthly backup but the day provided is \
missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a monthly backup but the day \
provided is missing""")

    def testCreateMonthly2(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        various dates and no hour"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,15,28])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        [1,15,28]), True, """We created a monthly backup but the day(s) \
provided is/are missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0), dt.time(0,0,0), dt.time(0,0,0)]), True, """We \
created a monthly backup but the day(s) provided is/are missing""")

    def testCreateMonthly3(self):
        """We create a new TaskSchedule object with TB == Monthly and just
        one date and one hour"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1], 
        HD=[dt.time(12,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], [1]), 
        True, """We created a monthly backup but the day(s) provided \
is/are missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(12,0,0)]), True, """We created a monthly backup but the \
day(s) provided is/are missing""")

    def testCreateMonthly4(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        various dates and various hours"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,15,28], 
        HD=[dt.time(12,0,0),
        dt.time(14,0,0), dt.time(18,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        [1, 15, 28]), True, """We created a monthly backup but the day(s) \
provided is/are missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(12,0,0), dt.time(14,0,0), dt.time(18,0,0)]), True, 
        """We created a monthly backup but the day(s) provided is/are \
missing""")

    def testCreateMonthly5(self):
        """We create a new TaskSchedule object with TB == Monthly and no 
        date and no hour"""
        new_tb = t.TemporalBasis(1)
        self.failUnlessRaises(t.TaskScheduleError, t.TaskSchedule, None,
        new_tb, None, DM=[], HD=[]) 
    
    def testCreateMonthly6(self):
        """We create a new TaskSchedule object with TB == Monthly and every
        date shortcut possible"""
        now = dt.date.today()
        last_day = now+relativedelta.relativedelta(months=+1, day=1)
        new_tb = t.TemporalBasis(1)
        _all_combinations = [1,7,14,21,28,last_day] 
        new_ts = t.TaskSchedule(None, new_tb, 'seven_days', 'first_day', 
        'last_day', DM=[], HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        _all_combinations), True, """We created a monthly backup but the 
day(s) provided is/are missing""")
    
    def testCreateMonthly7(self):
        """We create a new TaskSchedule object with TB == Monthly and last_day
        shortcut activated"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', DM=[], HD=[])
        self.failUnlessEqual(new_ts.periods['LD'], True, """We created a \
monthly backup but the day(s) provided is/are missing""")
    
    def testCreateMonthly8(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        seven_days shortcut activated"""
        new_tb = t.TemporalBasis(1)
        _all_combinations = [7,14,21,28]
        new_ts = t.TaskSchedule(None, new_tb, 'seven_days', DM=[], HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        _all_combinations), True, """We created a monthly backup but the 
day(s) provided is/are missing""")
    
    def testCreateMonthly9(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        first_day shortcut activated"""
        new_tb = t.TemporalBasis(1)
        _all_combinations = [1]
        new_ts = t.TaskSchedule(None, new_tb, 'first_day', DM=[], HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        _all_combinations), True, """We created a monthly backup but the 
day(s) provided is/are missing""")
    
    def testCreateMonthly10(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        a date and hours not using time class"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1], HD=[0])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        [1]), True, """We created a monthly backup but the day(s) \
provided is/are missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a monthly backup but the \
day(s) provided is/are missing""")
        
    def testCreateMonthly11(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        various dates (some of them impossible) and no hour"""
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[15,25,48,99], HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DM'], 
        [15,25]), True, """We created a monthly backup but the day(s) \
provided is/are missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a monthly backup but the \
day(s) provided is/are missing""")
     
    def testCreateMonthly12(self):
        """We create a new TaskSchedule object with TB == Monthly and 
        various dates (all of them impossible) and no hour"""
        new_tb = t.TemporalBasis(1)
        self.failUnlessRaises(t.TaskScheduleError, t.TaskSchedule, None,
        new_tb, None, DM=[45,65,78,109], HD=[])
     
    def testCreateWeekly1(self):
        """We create a new TaskSchedule object with TB == Weekly and just
        one date and no hour"""
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=[1])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DW'], [1]), 
        True, """We created a weekly backup but the day(s) provided is/are \
missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a weekly backup but the \
day(s) provided is/are missing""")

    def testCreateWeekly2(self):
        """We create a new TaskSchedule object with TB == Weekly and 
        various dates and no hour"""
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=[1,2,3])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DW'], [1,2,3]),
        True, """We created a weekly backup but the day provided is \
missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a weekly backup but the \
day(s) provided is/are missing""")

    def testCreateWeekly3(self):
        """We create a new TaskSchedule object with TB == Weekly and just
        one date and one hour"""
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=[1], 
        HD=[dt.time(9,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DW'], [1]), 
        True, """We created a weekly backup but the day provided is \
missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(9,0,0)]), True, """We created a weekly backup but the \
day(s) provided is/are missing""")

    def testCreateWeekly4(self):
        """We create a new TaskSchedule object with TB == Weekly and 
        various dates and various hours"""
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=[1, 2, 6, 7], 
        HD=[dt.time(9,0,0), dt.time(14,0,0), dt.time(20,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DW'], 
        [1, 2, 6]), True, """We created a weekly backup but the day \
provided is missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(9,0,0), dt.time(14,0,0), dt.time(20,0,0)]), True, 
        """We created a weekly backup but the day(s) provided is/are \
missing""")

    def testCreateWeekly5(self):
        """We create a new TaskSchedule object with TB == Weekly and no 
        date and no hour"""
        new_tb = t.TemporalBasis(2)
        self.failUnlessRaises(t.TaskScheduleError, t.TaskSchedule, None,
        new_tb, None, DW=[], HD=[]) 
    
    def testCreateWeekly6(self):
        """We create a new TaskSchedule object with TB == Weekly and no 
        date and no hour but with a Monthly date shorcut activated"""
        new_tb = t.TemporalBasis(2)
        self.failUnlessRaises(t.TaskScheduleError, t.TaskSchedule, None, 
        new_tb, 'seven_days', DW=[], HD=[]) 
    
    def testCreateWeekly7(self):
        """We create a new TaskSchedule object with TB == Weekly and 
        a date and hours not using time class"""
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=[1], HD=[0])
        self.failUnlessEqual(self.compareLists(new_ts.periods['DW'], 
        [1]), True, """We created a weekly backup but the day \
provided is missing""")
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a weekly backup but the \
day(s) provided is/are missing""")

    def testCreateWeekly8(self):
        """We create a new TaskSchedule object with TB == Weekly and 
        various dates (all of the impossible) and no hour"""
        new_tb = t.TemporalBasis(2)
        self.failUnlessRaises(t.TaskScheduleError, t.TaskSchedule, None,
        new_tb, None, DW=[11,13,45], HD=[])
    
    def testCreateDaily1(self):
        """We create a new TaskSchedule object with TB == Daily and no 
        hour"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a daily backup but the \
hour(s) provided is/are missing""")
    
    def testCreateDaily2(self):
        """We create a new TaskSchedule object with TB == Daily and one 
        hour"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[dt.time(9,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(9,0,0)]), True, """We created a daily backup but the \
hour(s) provided is/are missing""")
    
    def testCreateDaily3(self):
        """We create a new TaskSchedule object with TB == Daily and various
        hours"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[dt.time(0,0,0), 
        dt.time(9,0,0), dt.time(15,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0), dt.time(9,0,0), dt.time(15,0,0)]), True, 
        """We created a daily backup but the hour(s) provided is/are \
missing""")
    
    def testCreateDaily4(self):
        """We create a new TaskSchedule object with TB == Daily and hours
        not using Time class"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[9,16,19])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a daily backup but the \
hour(s) provided is/are missing""")
    
    def testCreateDaily5(self):
        """We create a new TaskSchedule object with TB == Daily and  
         no hour but with a Monthly date shorcut activated"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a daily backup but the \
hour(s) provided is/are missing""")
    
    def testCreateDaily6(self):
        """We create a new TaskSchedule object with TB == Daily and  
         various dates and various hours"""
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,2,3,5], 
        HD=[dt.time(3,0,0), dt.time(8,0,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(3,0,0), dt.time(8,0,0)]), True, """We created a daily \
backup but the hour(s) provided is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, """We \
created a daily backup but the hour(s) provided is/are missing""")
    
    def testCreateHourly1(self):
        """We create a new TaskSchedule object with TB == Hourly and not
        using hours"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created an hourly backup but the \
hour(s) provided is/are missing""")
    
    def testCreateHourly2(self):
        """We create a new TaskSchedule object with TB == Hourly and using
        one hour"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[dt.time(7,15,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,15,0)]), True, """We created an hourly backup but the \
hour(s) provided is/are missing""")
    
    def testCreateHourly3(self):
        """We create a new TaskSchedule object with TB == Hourly and using
        various hours"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[dt.time(7,15,0), 
        dt.time(6,20,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,15,0), dt.time(0,20,0)]), True, """We created an hourly /
backup but the hour(s) provided is/are missing""")
    
    def testCreateHourly4(self):
        """We create a new TaskSchedule object with TB == Hourly and not using
        Time class"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, HD=[(9,0,0), (16,15,0), 
        (19,50,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created a daily backup but the \
hour(s) provided is/are missing""")
    
    def testCreateHourly5(self):
        """We create a new TaskSchedule object with TB == Hourly and  
         no hour but with a Monthly date shorcut activated"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', HD=[])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0)]), True, """We created an hourly backup but the \
hour(s) provided is/are missing""")
    
    def testCreateHourly6(self):
        """We create a new TaskSchedule object with TB == Hourly and  
         various dates and various hours"""
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,2,3,5], 
        HD=[dt.time(3,0,0), dt.time(8,40,0)])
        self.failUnlessEqual(self.compareLists(new_ts.periods['HD'], 
        [dt.time(0,0,0), dt.time(0,40,0)]), True, """We created an hourly \
backup but the hour(s) provided is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, """We \
created hourly backup but the hour(s) provided is/are missing""")
   
    def testOnDemand1(self):
        """We create a new TaskSchedule object with TB == On Demand and
        no date and no hour"""
        new_tb = t.TemporalBasis(5)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[], DW=[], HD=[])
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DW'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")
        self.failUnlessEqual(new_ts.periods.has_key('HD'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")
 
    def testOnDemand2(self):
        """We create a new TaskSchedule object with TB == On Demand and
        various dates and various hours"""
        new_tb = t.TemporalBasis(5)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,8,10,45], 
        DW=[1,6,9], HD=[dt.time(5,8,9), dt.time(0,0,0), dt.time(13,8,9)])
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DW'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")
        self.failUnlessEqual(new_ts.periods.has_key('HD'), False, 
        """We created an "On Demand" backup but the hour(s) provided is/are \
missing""")

    def testInThisMoment1(self):
        """We create a new TaskSchedule object with TB == In This Moment and
        no date and no hour"""
        new_tb = t.TemporalBasis(6)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[], DW=[], HD=[])
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DW'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('HD'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")
 
    def testInThisMoment2(self):
        """We create a new TaskSchedule object with TB == In This Moment and
        various dates and various hours"""
        new_tb = t.TemporalBasis(6)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[1,8,10,45], 
        DW=[1,6,9], HD=[dt.time(5,8,9), dt.time(0,0,0), dt.time(13,8,9)])
        self.failUnlessEqual(new_ts.periods.has_key('DM'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('DW'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")
        self.failUnlessEqual(new_ts.periods.has_key('HD'), False, 
        """We created an "In This Moment" backup but the hour(s) provided \
is/are missing""")

    def testNextScheduledDate1(self):
        """Test this method in various random scenarios: Retrieving None"""
        new_tb = t.TemporalBasis(5)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[4,6,7,99], 
        HD=[dt.time(8,0)])
        gen_data = new_ts.next_scheduled_date()
        self.failUnless(gen_data is None, """We expected None as returned \
value. \nInstead we got %s""" % (gen_data))
  
    def testNextScheduledDate2(self):
        """Test this method in various random scenarios: Retrieving None"""
        new_tb = t.TemporalBasis(6)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=[4,6,7,99], 
        HD=[dt.time(8,0)])
        gen_data = new_ts.next_scheduled_date()
        self.failUnless(gen_data is None, """We expected None as returned \
value. \nInstead we got %s""" % (gen_data))
  
    def testNextScheduledDate3(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        one date"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DM': 20, 31 
        d_list = [20,31]
        # Expected output: 
        exp_dt = [dt.datetime(2008,03,20,0,0,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
  
    def testNextScheduledDate4(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        two dates"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DM': 20, 31 
        d_list = [20, 31]
        # Expected output: 
        exp_dt = [dt.datetime(2008,3,20,0,0,0),
                  dt.datetime(2008,3,31,0,0,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
  
    def testNextScheduledDate5(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        two dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DM': 5, 20, 31 
        d_list = [5, 20, 31]
        # Expected output: 
        exp_dt = [dt.datetime(2008,3,20,0,0,0),
                  dt.datetime(2008,3,31,0,0,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
  
    def testNextScheduledDate6(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        12 dates"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19, 12, 50)
        # Input values: 'DM': 19
        d_list = [19]
        # Expected output: 
        exp_dt = [dt.datetime(2008,4,19,0,0,0),
                  dt.datetime(2008,5,19,0,0,0),
                  dt.datetime(2008,6,19,0,0,0),
                  dt.datetime(2008,7,19,0,0,0),
                  dt.datetime(2008,8,19,0,0,0),
                  dt.datetime(2008,9,19,0,0,0),
                  dt.datetime(2008,10,19,0,0,0),
                  dt.datetime(2008,11,19,0,0,0),
                  dt.datetime(2008,12,19,0,0,0),
                  dt.datetime(2009,1,19,0,0,0),
                  dt.datetime(2009,2,19,0,0,0),
                  dt.datetime(2009,3,19,0,0,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 12)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
  
    def testNextScheduledDate7(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        12 dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DM': 19
        d_list = [19]
        # Expected output: 
        exp_dt = [dt.datetime(2008,3,19,11,0,0),
                  dt.datetime(2008,3,19,17,0,0),
                  dt.datetime(2008,4,19,11,0,0),
                  dt.datetime(2008,4,19,17,0,0),
                  dt.datetime(2008,5,19,11,0,0),
                  dt.datetime(2008,5,19,17,0,0),
                  dt.datetime(2008,6,19,11,0,0),
                  dt.datetime(2008,6,19,17,0,0),
                  dt.datetime(2008,7,19,11,0,0),
                  dt.datetime(2008,7,19,17,0,0),
                  dt.datetime(2008,8,19,11,0,0),
                  dt.datetime(2008,8,19,17,0,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, 
        HD=[dt.time(11,0), dt.time(17,0)])
        gen_data = new_ts.next_scheduled_date(cmp_dt, 12)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate8(self):
        """Test this method in various random scenarios: Retrieving Monthly 
        12 dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DM': 1 
        d_list = [1]
        h_list = [dt.time(0,0),
                  dt.time(0,15),
                  dt.time(0,30),
                  dt.time(0,45),
                  dt.time(5,15),
                  dt.time(5,30),
                  dt.time(5,45),
                  dt.time(8,15),
                  dt.time(8,30),
                  dt.time(9,45),
                  dt.time(10,15),
                  dt.time(11,30),
                  dt.time(12,45),
                  dt.time(22,15),
                  dt.time(23,30),
                  dt.time(23,45)]
        # Expected output: 
        exp_dt = [dt.datetime(2008,4,1,0,0,0),
                  dt.datetime(2008,4,1,0,15,0),
                  dt.datetime(2008,4,1,0,30,0),
                  dt.datetime(2008,4,1,0,45,0),
                  dt.datetime(2008,4,1,5,15,0),
                  dt.datetime(2008,4,1,5,30,0),
                  dt.datetime(2008,4,1,5,45,0),
                  dt.datetime(2008,4,1,8,15,0),
                  dt.datetime(2008,4,1,8,30,0),
                  dt.datetime(2008,4,1,9,45,0),
                  dt.datetime(2008,4,1,10,15,0),
                  dt.datetime(2008,4,1,11,30,0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 12)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate9(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        one date"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DW': 1
        d_list = [1] # tuesday
        h_list = []
        # expected output
        exp_dt = [dt.datetime(2008,3,25,0,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 1)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate10(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        two dates"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DW': 1
        d_list = [1] # tuesday
        h_list = []
        # expected output
        exp_dt = [dt.datetime(2008,3,25,0,0,0), dt.datetime(2008,4,1,0,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate11(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        two dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DW': 1 4
        d_list = [1, 4] 
        h_list = []
        # expected output
        exp_dt = [dt.datetime(2008,3,21,0,0,0), dt.datetime(2008,3,25,0,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate12(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        various dates"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008,03,19)
        # Input values: 'DW': 0, 1, 2, 3, 4, 5, 6
        d_list = [0, 1, 2, 3, 4, 5, 6] 
        h_list = []
        # expected output
        exp_dt = [dt.datetime(2008,3,19,0,0,0), dt.datetime(2008,3,20,0,0,0),
                  dt.datetime(2008,3,21,0,0,0), dt.datetime(2008,3,22,0,0,0),
                  dt.datetime(2008,3,23,0,0,0), dt.datetime(2008,3,24,0,0,0),
                  dt.datetime(2008,3,25,0,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 7)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate13(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        various dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 12, 1, 0)
        # Input values: 'DW': 0, 1, 2, 3, 4, 5, 6
        d_list = [0, 1, 2, 3, 4, 5, 6] 
        h_list = [dt.time(12,0,0), dt.time(14,0,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,19,14,0,0), 
                  dt.datetime(2008,3,20,12,0,0), 
                  dt.datetime(2008,3,20,14,0,0),
                  dt.datetime(2008,3,21,12,0,0),
                  dt.datetime(2008,3,21,14,0,0),
                  dt.datetime(2008,3,22,12,0,0),
                  dt.datetime(2008,3,22,14,0,0),
                  dt.datetime(2008,3,23,12,0,0),
                  dt.datetime(2008,3,23,14,0,0),
                  dt.datetime(2008,3,24,12,0,0),
                  dt.datetime(2008,3,24,14,0,0),
                  dt.datetime(2008,3,25,12,0,0),
                  dt.datetime(2008,3,25,14,0,0),
                  dt.datetime(2008,3,26,12,0,0),
                  dt.datetime(2008,3,26,14,0,0),
                  dt.datetime(2008,3,27,12,0,0),
                  dt.datetime(2008,3,27,14,0,0),
                  dt.datetime(2008,3,28,12,0,0),
                  dt.datetime(2008,3,28,14,0,0),
                  dt.datetime(2008,3,29,12,0,0),
                  dt.datetime(2008,3,29,14,0,0), 
                  dt.datetime(2008,3,30,12,0,0),
                  dt.datetime(2008,3,30,14,0,0),
                  dt.datetime(2008,3,31,12,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 24)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate14(self):
        """Test this method in various random scenarios: Retrieving Weekly 
        various dates (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19)
        # Input values: 'DW': 0, 1, 15
        d_list = [0, 1, 15] 
        h_list = [dt.time(12,0,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,24,12,0,0), 
                  dt.datetime(2008,3,25,12,0,0)]
        new_tb = t.TemporalBasis(2)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate15(self):
        """Test this method in various random scenarios: Retrieving Daily
        and one hour"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 55, 0)
        # Input values: 
        d_list = None
        h_list = [dt.time(12,0,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,20,12,0,0)]
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 1)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))

    def testNextScheduledDate16(self):
        """Test this method in various random scenarios: Retrieving Daily
        and one hour (variation)"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 56, 0)
        # Input values: 
        d_list = None
        h_list = [dt.time(14,55,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,20,14,55,0)]
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, DW=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 1)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate17(self):
        """Test this method in various random scenarios: Retrieving Daily
        and various hours"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 55, 0)
        # Input values: 
        d_list = [1,4,5,6,7]
        h_list = [dt.time(0,0,0), dt.time(1,0,0), dt.time(15,15,0),
                  dt.time(20,0,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,19,15,15,0), 
                  dt.datetime(2008,3,19,20,0,0),
                  dt.datetime(2008,3,20,0,0,0),
                  dt.datetime(2008,3,20,1,0,0),
                  dt.datetime(2008,3,20,15,15,0),
                  dt.datetime(2008,3,20,20,0,0),
                  dt.datetime(2008,3,21,0,0,0),
                  dt.datetime(2008,3,21,1,0,0),
                  dt.datetime(2008,3,21,15,15,0),
                  dt.datetime(2008,3,21,20,0,0)]
        new_tb = t.TemporalBasis(3)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 10)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate18(self):
        """Test this method in various random scenarios: Retrieving Hourly
        and one hour"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 55, 0)
        # Input values: 
        d_list = []
        h_list = [dt.time(0,0,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,19,15,0,0)] 
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 1)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
         
    def testNextScheduledDate19(self):
        """Test this method in various random scenarios: Retrieving Hourly
        and various hours"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 55, 0)
        # Input values: 
        d_list = []
        h_list = [dt.time(10,10,0), dt.time(14,55,0)]
        # expected output
        exp_dt = [dt.datetime(2008,3,19,14,55,0),
                  dt.datetime(2008,3,19,15,10,0)] 
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 2)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))

    def testNextScheduledDate20(self):
        """Test this method in various random scenarios: Retrieving Hourly
        and various hours"""
        # Comparing Date: 
        cmp_dt = dt.datetime(2008, 3, 19, 14, 55, 0)
        # Input values: 
        d_list = [4,6,99]
        h_list = [dt.time(0,12,0), 
                  dt.time(14,50,0),
                  dt.time(12,45,0),
                  dt.time(13,45,0),
                  dt.time(0,0,1),
                  dt.time(0,1,59)]
        # expected output
        exp_dt = [dt.datetime(2008,3,19,15,0,0),
                  dt.datetime(2008,3,19,15,1,0),
                  dt.datetime(2008,3,19,15,12,0),
                  dt.datetime(2008,3,19,15,45,0),
                  dt.datetime(2008,3,19,15,50,0),
                  dt.datetime(2008,3,19,16,0,0),
                  dt.datetime(2008,3,19,16,1,0),
                  dt.datetime(2008,3,19,16,12,0)] 
        new_tb = t.TemporalBasis(4)
        new_ts = t.TaskSchedule(None, new_tb, None, DM=d_list, HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 8)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))

    def testNextScheduledDate21(self):
        """Test this method in various random scenarios: Monthly with last_day
        shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 27, 15, 19, 0)
        # Input values
        d_list = [3, 10]
        h_list = [dt.time(12, 0 , 0)] 
        # expected output
        exp_dt = [dt.datetime(2008, 4, 1, 0, 0, 0),
                  dt.datetime(2008, 4, 3, 12, 0, 0),
                  dt.datetime(2008, 4, 10, 12, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 3)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate22(self):
        """Test this method in various random scenarios: Monthly with last_day
        shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 27, 15, 19, 0)
        # Input values
        d_list = [3, 10]
        h_list = [dt.time(12, 0 , 0), dt.time(18, 0, 0)] 
        # expected output
        exp_dt = [dt.datetime(2008, 4, 1, 0, 0, 0),
                  dt.datetime(2008, 4, 3, 12, 0, 0),
                  dt.datetime(2008, 4, 3, 18, 0, 0),
                  dt.datetime(2008, 4, 10, 12, 0, 0),
                  dt.datetime(2008, 4, 10, 18, 0, 0),
                  dt.datetime(2008, 5, 1, 0, 0, 0),
                  dt.datetime(2008, 5, 3, 12, 0, 0),
                  dt.datetime(2008, 5, 3, 18, 0, 0),
                  dt.datetime(2008, 5, 10, 12, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 9)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate23(self):
        """Test this method in various random scenarios: Monthly with last_day
        shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 27, 15, 19, 0)
        # Input values
        d_list = [1, 30]
        h_list = [dt.time(0, 0 , 0), dt.time(0, 15, 0), dt.time(12, 0, 0)] 
        # expected output
        exp_dt = [dt.datetime(2008, 3, 30, 0, 0, 0),
                  dt.datetime(2008, 3, 30, 0, 15, 0),
                  dt.datetime(2008, 3, 30, 12, 0, 0),
                  dt.datetime(2008, 4, 1, 0, 0, 0),
                  dt.datetime(2008, 4, 1, 0, 15, 0),
                  dt.datetime(2008, 4, 1, 12, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 6)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))

    def testNextScheduledDate24(self):
        """Test this method in various random scenarios: Monthly with last_day
        shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 27, 15, 19, 0)
        # Input values
        d_list = None
        h_list = None 
        # expected output
        exp_dt = [dt.datetime(2008, 4, 1, 0, 0, 0),
                  dt.datetime(2008, 5, 1, 0, 0, 0),
                  dt.datetime(2008, 6, 1, 0, 0, 0),
                  dt.datetime(2008, 7, 1, 0, 0, 0),
                  dt.datetime(2008, 8, 1, 0, 0, 0),
                  dt.datetime(2008, 9, 1, 0, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'last_day', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 6)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate25(self):
        """Test this method in various random scenarios: Monthly with 
        seven_days shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 30, 20, 35, 0)
        # Input values
        d_list = None
        h_list = None 
        # expected output
        exp_dt = [dt.datetime(2008, 4, 7, 0, 0, 0),
                  dt.datetime(2008, 4, 14, 0, 0, 0),
                  dt.datetime(2008, 4, 21, 0, 0, 0),
                  dt.datetime(2008, 4, 28, 0, 0, 0),
                  dt.datetime(2008, 5, 7, 0, 0, 0),
                  dt.datetime(2008, 5, 14, 0, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'seven_days', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 6)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))

    def testNextScheduledDate26(self):
        """Test this method in various random scenarios: Monthly with 
        firt_day shortcut activated"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 30, 20, 35, 0)
        # Input values
        d_list = None
        h_list = None 
        # expected output
        exp_dt = [dt.datetime(2008, 4, 1, 0, 0, 0),
                  dt.datetime(2008, 5, 1, 0, 0, 0),
                  dt.datetime(2008, 6, 1, 0, 0, 0),
                  dt.datetime(2008, 7, 1, 0, 0, 0),
                  dt.datetime(2008, 8, 1, 0, 0, 0),
                  dt.datetime(2008, 9, 1, 0, 0, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'first_day', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 6)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
    def testNextScheduledDate27(self):
        """Test this method in various random scenarios: Monthly with 
        seven_days shortcut activated and 2 hours"""
        # Comparing Date:
        cmp_dt = dt.datetime(2008, 3, 30, 20, 35, 0)
        # Input values
        d_list = None
        h_list = [dt.time(13,45), dt.time(19,55)]
        # expected output
        exp_dt = [dt.datetime(2008, 4, 7, 13, 45, 0),
                  dt.datetime(2008, 4, 7, 19, 55, 0),
                  dt.datetime(2008, 4, 14, 13, 45, 0),
                  dt.datetime(2008, 4, 14, 19, 55, 0),
                  dt.datetime(2008, 4, 21, 13, 45, 0),
                  dt.datetime(2008, 4, 21, 19, 55, 0)]
        new_tb = t.TemporalBasis(1)
        new_ts = t.TaskSchedule(None, new_tb, 'seven_days', DM=d_list, 
        HD=h_list)
        gen_data = new_ts.next_scheduled_date(cmp_dt, 6)
        self.failUnless(gen_data==exp_dt, """We expected %s as returned \
value. \nInstead we got %s""" % (exp_dt, gen_data))
        
def SuiteTS():
    suite = unittest.TestLoader().loadTestsFromTestCase(testTaskSchedule)
    return suite

if __name__ == '__main__':
    sqlcon.cursor.execute("""delete from watched_file""")
    sqlcon.cursor.execute("""delete from task_schedule""")
    sqlcon.cursor.execute("""delete from task""")
    sqlcon.commit()
    runner = unittest.TextTestRunner()
    loader = unittest.TestLoader()
    results = unittest.TestResult()
    suite = SuiteTS()
    coverage.erase()
    coverage.start()
    results = runner.run(suite)
    coverage.stop()
    coverage.report('../task.py')
    coverage.erase()
    sqlcon.cursor.execute("""delete from task""")
    sqlcon.commit()

