#! /usr/bin/env python
# Copyright (C) 2008 Martin Albisetti

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""This module manages the inotify watcher"""

# standard library imports
import gettext
from datetime import datetime
import time
import itertools
import os

# related 3rd party imports

# local application/ library specific imports.
import pyinotify
import createdb
from logger import *

# establish translation context
gettext.bindtextdomain('Ghosts', 'locale')
gettext.textdomain('Ghosts')
_ = gettext.gettext

__all__ = ['InotifyDescriptor', 'InotifyError', 'NEvent']

# connects with the database
# we default to 20 seconds of timeout to give room to other changes.
sqlcon = createdb.SQLConnection(20)
# events to watch
NEvents =  ['IN_ATTRIB', # metadata changed, we need to update info
            'IN_DELETE', # a file was deleted, we need to take it out
            'IN_CLOSE_WRITE', # a created file was closed (we need to tie
                              # this event to IN_CREATE) so we can backup it
            #'IN_ONESHOT', # report file just once
            'IN_MODIFY', # a file was modified, we need to backup it
            'IN_MOVE_SELF', # a file was moved (see below)
            'IN_MOVED_FROM', # a file came here, we might need to backup it
            'IN_MOVED_TO'] # a file got out, we need to take it out

def _calculate_mask(events):
    """Given a sequence of pyinotify events this function calculates the
    mask and returns it as an integer"""
    return reduce(lambda x, y: x|y, [getattr(pyinotify, x) for x in events])

def _is_parent(parent, child):
    """Given two paths this func. will return true if the second one is child
    of the first one"""
    for idx, val in enumerate(parent.split("/")):
        if val != child.split("/")[idx]:
            return False
    return True

class InotifyError(Exception):
    """A generic error class for Database Errors"""
    pass

class WatchManagerError(pyinotify.WatchManagerError):
    pass

class WatchManager(pyinotify.WatchManager):
    """
    Provide operations for watching files and directories. Integrated
    dictionary is used to reference watched items.
    """

    def add_watch(self, path, mask, proc_fun=None, rec=False,
                  auto_add=False, do_glob=False, quiet=True,
                  exclude_filter=None, task=None):
        """
        Add watch(s) on given path(s) with the specified mask and
        optionnally with a processing function and recursive flag.

        @param path: Path to watch, the path can either be a file or a
                     directory. Also accepts a sequence (list) of paths.
        @type path: string or list of string
        @param mask: Bitmask of events.
        @type mask: int
        @param proc_fun: Processing object.
        @type proc_fun: function or ProcessEvent instance or instance of
                        one of its subclasses or callable object.
        @param rec: Recursively add watches from path on all its
                    subdirectories, set to False by default (doesn't
                    follows symlinks).
        @type rec: bool
        @param auto_add: Automatically add watches on newly created
                         directories in the watch's path.
        @type auto_add: bool
        @param do_glob: Do globbing on pathname.
        @type do_glob: bool
        @param quiet: if True raise an WatchManagerError exception on
                      error. See example not_quiet.py
        @type quiet: bool
        @param exclude_filter: boolean function, returns True if current
                               path must be excluded from being watched.
                               Has precedence on exclude_filter defined
                               into __init__.
        @type exclude_filter: bool
        @param task: for each path the task identification number
        @type task: list
        @return: dict of paths associated to watch descriptors. A wd value
                 is positive if the watch has been sucessfully added,
                 otherwise the value is negative. If the path is invalid
                 it will be not included into this dict.
        @rtype: dict of str: [int, tuple(int. int), [int, ...]]
        """
        is_root = False
        parent_wd = None
        v_par = None
        ret_ = {} # return {path: wd, [parent wd, creation order], [task(s), ...}

        if exclude_filter is None:
            exclude_filter = self._exclude_filter

        # This routine builds a dict for every dir selected (and its
        # subdirectories in case "rec" is True) and put them on watch, calling
        # __add_watch. A user might select a given path more than once to make
        # it backup (don't ask me why) so we want to prevent calling the
        # inotify lib for an already watched dir. That's whi we call
        # ret_.has_key to see if it's already exists.
        # Original pyinotify returns a dict composed by path:wd. We needed to
        # incorporate more info in this structure to help us to track subdirs
        # (since they generate a new wd) and to aux. us in this functionality
        # of not asking the lib to watch the path again.
        # For that, we build the dict with the following structure:
        #       key: path (as usual)
        #       value: [wd, tuple(parent wd*, order**), [task id***]]
        #       where *: the original path selected to be watch, the rest of
        #       them are subdirs.
        #            **: the order it got when the path tree was walked
        #           ***: task: a list of all tasks that are interested in the
        #           given path

        # normalize args as list elements
        for idx1, npath in enumerate(self.__format_param(path)):
            # a previous task already asked to watch this path, so we don't
            # need to make it watch again, we just need to inherit the assigned
            # wd.
            # TODO: this performs an equality comparison but if the new path is
            # parent of an existing one we might now catch it
            if ret_.has_key(npath):
                v_par = ret_[npath][1]
                for outer in [inner for inner in ret_.iteritems() \
                        if inner[1][1][0] == v_par[0] and \
                           inner[1][1][1] >= v_par[1]]:
                    ret_[outer[0]][2].append(task[idx1])
                continue
            else:
                # unix pathname pattern expansion
                for apath in self.__glob(npath, do_glob):
                    # recursively list subdirs according to rec param
                    for idx2, rpath in enumerate(self.__walk_rec(apath, rec)):
                        # if this is the root dict, save the generated wd to
                        # use it later as auxiliary value
                        if rpath == npath:
                            is_root = True
                            parent_wd = None
                        else:
                            is_root = False
                        if not exclude_filter(rpath):
                            wd = self.__add_watch(rpath, mask, proc_fun,
                                    auto_add)
                            if wd < 0:
                                err = 'add_watch: cannot watch %s (WD=%d)'
                                err = err % (rpath, wd)
                                if quiet:
                                    log.error(err)
                                else:
                                    raise WatchManagerError(err, ret_)
                            else:
                                if is_root:
                                    parent_wd = wd
                                ret_[rpath] = [wd, (parent_wd, idx2),
                                        [task[idx1]]]
                        else:
                            # Let's say -w means 'explicitely excluded
                            # from watching'.
                            ret_[rpath] = [-2, (parent_wd, idx2), [task[idx1]]]
        return ret_

    def __walk_rec(self, top, rec):
        """
        Yields each subdirectories of top, doesn't follow symlinks.
        If rec is false, only yield top.

        @param top: root directory.
        @type top: string
        @param rec: recursive flag.
        @type rec: bool
        @return: path of one subdirectory.
        @rtype: string
        """
        if not rec or os.path.islink(top) or not os.path.isdir(top):
            yield unicode(top)
        else:
            for root, dirs, files in os.walk(top):
                yield unicode(root)


class InotifyDescriptor(object):
    """This class represents all inotify watch descriptors created for
    the current session: it tracks the task id and path related and gets
    notifications of any new task created during the lifetime of the session
    """

    __table__ = "inotify_descriptor"
    # special flags to notify the daemon of certaing activities that otherwise
    # would be hard to make it notice. These flags maps inotify.ind_flag
    stop_flag = 0
    delete_flag = 1

    def __init__(self, SQLCon=None):
        """If SQLCon (as instance of createdb.SQLConnection) exists, use this
        connection handler and ignore the one provided by this module"""
        if not SQLCon:
            self.sqlcon = sqlcon
        else:
            self.sqlcon = SQLCon
        self.rc = None

    def find(self):
        """Returns a list of dicts containing new path and action flag
        These values cannot have a wd assigned (unless the task is
        canceled)"""
        self.rc = self.sqlcon.find(self, ("ind_path","ind_flag","ind_wd_id",
        "ind_tas_id"), dict(ind_wd_id=None,SQL=" or ind_flag=1"))

    def update(self, wd_dict):
        """Update descriptors that were assigned a wd.
        Parameters:
            wd_dict: a dictionary containing 'path watched':['wd', [reserved],
            [task(s)]]
        """
        wd_list = []
        for rs1 in wd_dict.iteritems():
            for rs2 in rs1[1][2]:
                wd_list.append(dict(ind_tas_id=rs2,ind_path=rs1[0],
                    ind_wd_id=rs1[1][0],ind_timestamp=datetime.now()))
            self.sqlcon.insert(self, wd_list, 'replace', False)

    def insert(self, tas_id, tas_from):
        """A new task has been created so there's a new path to watch."""
        self.sqlcon.insert(self, dict(ind_tas_id=tas_id, ind_path=tas_from,
        ind_timestamp=datetime.now()), 'abort', False)

    def restart(self):
        """When required, this function returns a list of tasks that need to
        backup a.s.a. there are changes. We need to reload the watch class
        with all the tasks with the required information.

        No input parameters. Returns a list containing dicts, each one
        containing the to-backup path (Task.tas_from) and the task id
        """
        self.sqlcon.delete(self)
        # Preventing a circular import due to task.Task
        self.sqlcon.cursor.execute("""select tas_from, tas_id from task where
        tas_cancel_date is null and tas_teb_id = 6""")
        self.rc = self.sqlcon.cursor.fetchall()
        for cur in self.rc:
            self.insert(cur[1], cur[0])

    def stop(self):
        """In case of need, user can make the watcher stop smoothly. This
        method will update inotify_descriptor.ind_flag to STOP and the
        daemon -as soon as it read it- will stop.
        """
        self.sqlcon.insert(self, dict(ind_tas_id=0, ind_path=u"/dev",
        ind_timestamp=datetime.now(), ind_flag=self.stop_flag), 'abort', False)

    def delete(self, tas_id=None, wd=None):
        """Delete a given task from inotify watcher. If tas_id is null,
        wd must be a dict and the wds contained there will be erased from
        the database"""
        if tas_id:
            self.sqlcon.update(self, {"ind_flag":self.delete_flag},
                                     {"ind_tas_id":tas_id}, False)
        if wd:
            self.sqlcon.delete(self, dict(ind_wd_id=[y for (y,z) in \
                                          wd.iteritems() if z is True],
                                          ind_flag=self.delete_flag), False)
            self.sqlcon.update(self, {"ind_wd_id":None},
                                     dict(ind_wd_id=[y for (y,z) in \
                                     wd.iteritems() if z is True],
                                     ind_flag=None), False)

class NEvent(object):
    """This class represents all generated events from the watcher"""

    __table__ = "event"
    __pk__ = "eve_seq"
    eve_tas_id = None
    eve_seq = None

    def __init__(self, id=None, **kwargs):
        """If id is not null, get the given NEvent from the database.
        If kwargs is null, init an empty instance (must be fill later with db
        data). kwargs must map to columns in event table. If key "eve_tas_id"
        is missing or None from kwargs, it will be retrieved from the database.
        If it does not exist in there, the init will fail"""
        if id:
            rc = sqlcon.get(self, id)
            for x in rc.items():
                setattr(self, x[0], x[1])
        else:
            for val in kwargs.iteritems():
                setattr(self, val[0], val[1])
            if not hasattr(self, "eve_timestamp"):
                self.eve_timestamp = datetime.now()
            if not self.eve_tas_id and kwargs:
                rc = sqlcon.find(InotifyDescriptor, ("ind_tas_id",),
                dict(ind_wd_id=self.eve_wd))
                if rc:
                    setattr(self, "eve_tas_id", rc[0]['ind_tas_id'])
                else:
                    log.error("""No task could be found with the given watch \
descriptor (%d)""" % (self.eve_wd))

    def save(self):
        """Save the given instance in the database: only for new events"""
        if not self.eve_seq:
            sqlcon.insert(self, dict(x for x in self.__dict__.iteritems() \
                    if not x[0].startswith('_')), 'abort')

    @staticmethod
    def pop(commit_=False, *min_max_eve):
        """Return the existing records in the database, deleting them from
        table. This method will return a list of instances each of them
        representing a NEvent.
        Parameters:
        commit: If True it will commit the transaction, False otherwise
        min_max_eve: minimum and maximum event to retrieve and erase
        """
        if min_max_eve:
            events = sqlcon.find(NEvent, {"eve_tas_id":"eve_tas_id",
            "eve_path":"eve_path", "eve_name":"eve_name"},
            dict(eve_tas_id='!None', SQL=""" and eve_seq between %d and %d
            group by eve_tas_id, eve_path, eve_name """ % (min_max_eve[0],
            min_max_eve[1])), True)
            low_id = min_max_eve[0]
            upp_id = min_max_eve[1]
        else:
            events = sqlcon.find(NEvent, None, None, True)
            low_id = events[0].eve_seq
            upp_id = events[-1].eve_seq
        sqlcon.delete(NEvent, dict(SQL=" eve_seq between %d and %d" % (low_id,
            upp_id)), commit_)
        return events

class NEvent(object):
    """This class represents all generated events from the watcher"""

    __table__ = "event"
    __pk__ = "eve_seq"
    eve_tas_id = None

    def __init__(self, id=None, **kwargs):
        """If id is null and no kwargs dict is supplied, will instantiate an
        empty object.
        If id is not null, get the given NEvent from the database.
        If kwargs is null, init an empty instance (must be fill later with db
        data). kwargs must map to columns in event table. If key "eve_tas_id"
        is missing or None from kwargs, it will be retrieved from the database.
        If it does not exist in there, the init will fail"""
        if id:
            rc = sqlcon.get(self)
            for x in rc.items():
                setattr(self, x[0], x[1])
        else:
            for val in kwargs.iteritems():
                setattr(self, val[0], val[1])
            self.eve_timestamp = datetime.now()
            if not self.eve_tas_id and kwargs:
                rc = sqlcon.find(InotifyDescriptor, ("ind_tas_id",),
                                 dict(ind_wd_id=self.eve_wd))
                if rc:
                    setattr(self, "eve_tas_id", rc[0])
                else:
                    raise InotifyError(_("""No task could be found with the \
given watch descriptor (%d)""" % (self.eve_wd)))

    def save(self):
        """Save the given instance in the database: only for new events"""
        if not self.eve_seq:
            sqlcon.insert(self, dict(x for x in self.__dict__.iteritems() \
                    if not x[0].startswith('_')), 'abort')

    @staticmethod
    def pop(commit_=False):
        """Return the existing records in the database, deleting them from
        table. This method will return a list of instances each of them
        representing a NEvent."""
        events = sqlcon.find(NEvent, None, None, True)
        sqlcon.delete(NEvent, dict(eve_seq=[x.eve_seq for x in events]),
                commit_)
        return events

class NProcess(pyinotify.ProcessEvent):
    """Define all required methods to handle actions regarding events
    happening in a watched directory"""

    def process_default(self, eve):
        # we actually exclude files with forbidden extensions
        # TODO: this "exclusion" does not seem to be working. Find out why
        if len([val for val in itertools.takewhile(eve.name.endswith,
            excluded_extensions)])==0:
            # instantiate NEvent to insert the data in the db
            nevent = NEvent(None, **dict(("eve_%s" % (x),y) for (x, y) in \
                            eve.__dict__.iteritems()))
            nevent.save()

    def process_IN_CLOSE_NOWRITE(self, eve):
        """In case an unwrittable object is closed and this object is a dir,
        we ignore the event"""
        if eve.dir:
            pass
        else:
            self.process_default(eve)

def cb_events(inst):
    """Callback method to update the watcher's list with newly added tasks and
    removed ones"""
    # ** BEGIN RETRIEVE SECTION **
    ID.find()
    # ** END RETRIEVE SECTION **

    # ** BEGIN STOP SECTION **
    # the keyword STOP was found in the list: we stop the daemon
    if ID.stop_flag in [x["ind_flag"] for x in ID.rc]:
        raise KeyboardInterrupt
    # ** END STOP SECTION **

    # ** BEGIN ADD SECTION **
    wd = WM.add_watch([x["ind_path"] for x in ID.rc if \
                      x["ind_flag"] is None],
                      _calculate_mask(NEvents), quiet=False, rec=True,
                      auto_add=True, exclude_filter=None,
                      task=[x["ind_tas_id"] for x in ID.rc])
    ID.update(wd)
    sqlcon.commit()
    # ** END ADD SECTION **

    # ** BEGIN REMOVE SECTION **
    # if we have to, remove watches
    wd = WM.rm_watch([x["ind_wd_id"] for x in ID.rc if \
                     x["ind_flag"] == ID.delete_flag], True)
    ID.delete(None, wd)
    sqlcon.commit()
    # ** END REMOVE SECTION **

if __name__ == '__main__':
    # coded sql to prevent cyclical import
    # we find out all globally excluded extensions
    sqlcon.cursor.execute("""select glp_value from global_parameter where
    glp_id='global_excluded_extension'""")
    excluded_extensions = sqlcon.cursor.fetchone()[0]
    ID = InotifyDescriptor()
    WM = WatchManager()
    notifier = pyinotify.Notifier(WM, NProcess(), timeout=20)
    ID.restart()
    wd = WM.add_watch([x[0] for x in ID.rc], _calculate_mask(NEvents),
                      rec=True, auto_add=True, exclude_filter=None,
                      task=[x[1] for x in ID.rc])
    ID.update(wd)
    sqlcon.commit()
    notifier.loop(cb_events)
